# -*- coding: utf-8 -*-

from simmate.database.base_data_types import Structure, Thermodynamics, table_column


class MatprojStructure(Structure, Thermodynamics):
    """
    Crystal structures from the [Materials Project](https://materialsproject.org/)
    database.

    Currently, this table only stores strucure and thermodynamic information,
    but the Materials Project has much more data available via their
    [REST API](https://github.com/materialsproject/api) and website.
    """

    class Meta:
        app_label = "third_parties"

    archive_fields = [
        "energy_uncorrected",
        "band_gap",
        "is_gap_direct",
        "is_magnetic",
        "total_magnetization",
        "is_theoretical",
    ]
    api_filters = dict(
        energy_uncorrected=["range"],
        band_gap=["range"],
        is_gap_direct=["exact"],
        is_magnetic=["exact"],
        total_magnetization=["range"],
        is_theoretical=["exact"],
    )
    source = "Materials Project"
    source_long = "The Materials Project at Berkeley National Labs"
    homepage = "https://materialsproject.org/"
    source_doi = "https://doi.org/10.1063/1.4812323"
    remote_archive_link = "https://archives.simmate.org/MatprojStructure-2022-08-27.zip"

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "mp-12345")
    """

    energy_uncorrected = table_column.FloatField(blank=True, null=True)
    """
    The reported energy of the system BEFORE Materials Project applies
    their composition-based corrections.
    """

    band_gap = table_column.FloatField(blank=True, null=True)
    """
    The band gap energy in eV.
    """

    is_gap_direct = table_column.BooleanField(blank=True, null=True)
    """
    Whether the band gap is direct or indirect.
    """

    is_magnetic = table_column.BooleanField(blank=True, null=True)
    """
    Whether the material is magnetic
    """

    total_magnetization = table_column.FloatField(blank=True, null=True)
    """
    The total magnetization of the material
    """

    is_theoretical = table_column.BooleanField(blank=True, null=True)
    """
    Whether the material is from a theoretical structure. False indicates
    that it is experimentally known.
    """

    updated_at = table_column.DateTimeField(blank=True, null=True)
    """
    Timestamp of when this row was was lasted changed / updated by the 
    Materials Project
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the Materials Project website.
        """
        # All Materials Project structures have their data mapped to a URL in
        # the same way. For example...
        #   https://materialsproject.org/materials/mp-12345/
        return f"https://materialsproject.org/materials/{self.id}/"



# -*- coding: utf-8 -*-

from pymatgen.transformations.standard_transformations import (
    PerturbStructureTransformation,
)

from simmate.toolkit import Structure
from simmate.toolkit.transformations.base import Transformation


class CoordinatePerturbation(Transformation):
    """
    Randomly perturbs or "rattles" atomic sites within a structure. This
    transformation is almost guaranteed to give a structure with P1 symmetry.

    This is a wrapper around pymatgen's
    [PerturbStructureTransformation](https://pymatgen.org/pymatgen.transformations.standard_transformations.html#pymatgen.transformations.standard_transformations.PerturbStructureTransformation)

    NOTE: Random perturbation of sites was removed from USPEX because primarily made
    more defective + lower energy structures
    https://uspex-team.org/static/file/USPEX-LargeComplexSystems-2010.pdf

    NOTE: the function Structure.pertrub displaces each site at an equal distance
    this function on the other hand displaces at a variable distance selected
    from a guassian distribution between min/mix
    """

    io_scale = "one_to_one"
    ninput = 1
    allow_parallel = False

    @staticmethod
    def apply_transformation(
        structure: Structure,
        min_displace: float = 0,
        max_displace: float = 1,
    ):

        perturb_object = PerturbStructureTransformation(min_displace, max_displace)

        new_structure = perturb_object.apply_transformation(structure)

        return new_structure



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.error_handlers import Zpotrf
from simmate.calculators.vasp.inputs import Incar
from simmate.conftest import copy_test_files, make_dummy_files


def test_eddrmm(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="zpotrf",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = tmp_path / "INCAR"
    chgcar_filename = tmp_path / "CHGCAR"
    wavecar_filename = tmp_path / "WAVECAR"
    oszicar_filename = tmp_path / "OSZICAR"

    # init class with default settings
    error_handler = Zpotrf()

    # Confirm an error IS NOT found
    error_handler.filename_to_check = "vasp.no_error"
    assert error_handler.check(tmp_path) == False

    # Confirm an error IS found
    error_handler.filename_to_check = "vasp.out"
    assert error_handler.check(tmp_path) == True

    # Attempt to fix the error when we have >1 valid ionic steps in the OSZICAR
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmp_path)
    assert (
        fix
        == "set ISYM to 0 and switched POTIM from 0.5 to 0.25 and deleted CHGCAR and WAVECAR"
    )
    incar = Incar.from_file(incar_filename)
    assert incar["ISYM"] == 0
    assert incar["POTIM"] == 0.25
    assert not chgcar_filename.exists()
    assert not wavecar_filename.exists()

    # remove OSZICAR to the remaining checks/fixes
    oszicar_filename.unlink()

    # Make first attempt at fixing the error
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmp_path)
    assert fix == "set ISYM to 0 and deleted CHGCAR and WAVECAR"
    incar = Incar.from_file(incar_filename)
    assert Incar.from_file(incar_filename)["ISYM"] == 0
    assert not chgcar_filename.exists()
    assert not wavecar_filename.exists()

    # Make second attempt at fixing the error
    with incar_filename.open("w") as file:
        file.write("NSW = 1 \nISIF = 2")
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmp_path)
    assert fix == "scaled the structure lattice by +20% and deleted CHGCAR and WAVECAR"



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.relaxation.matproj import (
    Relaxation__Vasp__Matproj,
)


class Relaxation__Vasp__MatprojMetal(Relaxation__Vasp__Matproj):
    """
    This task is a reimplementation of pymatgen's
    [MPMetalRelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPMetalRelaxSet).

    Runs a VASP relaxation calculation using MIT Project settings, where some
    settings are adjusted to accomodate metals. This include a denser kpt grid
    and proper smearing.
    """

    description_doc_short = "based on pymatgen's MPMetalRelaxSet"

    incar = Relaxation__Vasp__Matproj.incar.copy()
    incar.update(
        dict(
            ISMEAR=1,
            SIGMA=0.2,
            KSPACING=0.3,
        )
    )



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.diffusion.neb_from_images_mit import (
    Diffusion__Vasp__NebFromImagesMit,
)
from simmate.calculators.vasp.workflows.diffusion.neb_single_path_base import (
    SinglePathWorkflow,
)
from simmate.calculators.vasp.workflows.relaxation.mvl_neb_endpoint import (
    Relaxation__Vasp__MvlNebEndpoint,
)


class Diffusion__Vasp__NebSinglePathMit(SinglePathWorkflow):
    """
    Runs a full diffusion analysis on a single migration hop using NEB with
    MIT settings.
    """

    endpoint_relaxation_workflow = Relaxation__Vasp__MvlNebEndpoint
    from_images_workflow = Diffusion__Vasp__NebFromImagesMit



# -*- coding: utf-8 -*-

import logging

from numpy.random import choice, randint
from pymatgen.analysis.structure_matcher import StructureMatcher

from simmate.toolkit.transformations.base import Transformation


class AtomicPermutation(Transformation):
    """
    Two atoms of different types are exchanged a variable number of times
    See: https://uspex-team.org/static/file/JChemPhys-USPEX-2006.pdf
    """

    io_scale = "one_to_one"
    ninput = 1
    allow_parallel = False

    #!!! should I add an option for distribution of exchange number? Potentially
    # use creation.vector objects?
    # USPEX states that they do this - and select from a normal distribution

    @staticmethod
    def apply_transformation(
        structure,
        min_exchanges=1,
        max_exchanges=5,
        max_attempts=100,
    ):

        # grab a list of the elements
        elements = structure.composition.elements

        # This mutation is not possible for structures that have only one element
        if len(elements) == 1:
            print(
                "You cannot perform an atomic permutation on structure that "
                "only has one element type!!"
            )
            return False
        #!!! add another elif for when all sites are equivalent and atomic
        # permutation cannot create a new structure

        #!!! when doing multiple exchanges, I do not check to see if an exchange
        # has been done before - I can change this though
        # Therefore one exchange could undo another and we could be back where
        # we started. There are also scenarios where all sites are equivalent
        # and atomic permutation cannot create a new structure. An example of
        # this is NaCl, where an exchange still yields an identical structure
        # This leaves a chance that we end up with an identical structure to
        # what we started with. Therefore, I must hava a structurematcher object
        # that I use to ensure we have a new structure. We try making a new
        # structure X number of times (see max_attempts above) and if we
        # can't, we failed to mutate the structure.
        structure_matcher = StructureMatcher()

        for attempt in range(max_attempts):

            # Make a deepcopy of the structure so that we aren't modifying
            # it inplace. This also allows us to compare the new structure to
            # the original placing this at the top of the loop also resets
            # the structure for us each time
            new_structure = structure.copy()

            # grab a random integer within the exchange min/max defined above
            nexchanges = randint(low=min_exchanges, high=max_exchanges)

            # perform an exchange of two random atom types X number of times
            for n in range(nexchanges):
                # grab two elements of different types
                element1, element2 = choice(elements, size=2)

                # select a random index of element1 and element2
                index1 = choice(new_structure.indices_from_symbol(element1.symbol))
                index2 = choice(new_structure.indices_from_symbol(element2.symbol))

                # now exchange the species type of those two sites
                new_structure.replace(index1, element2)
                new_structure.replace(index2, element1)

            # see if the new structure is different from the original!
            # check will be True if the new structure is the same as the original
            check = structure_matcher.fit(structure, new_structure)
            if not check:
                # we successfully make a new structure!
                return new_structure
            # else continue

        # if we make it this far, then we hit our max_attempts limit without
        # making a new structure therefore, we failed the mutation
        logging.warning(
            "Failed to make a new structure that is different from the original"
        )
        return False



# -*- coding: utf-8 -*-

from simmate.toolkit.validators.fingerprint import FingerprintValidator

from .featurizers import RadialDistributionFunction as rdf


class RdfFingerprint(FingerprintValidator):

    comparison_mode = "cos"

    @staticmethod
    def get_featurizer(
        cutoff=20.0,
        bin_size=0.1,
        **kwargs,
    ):
        featurizer = rdf(cutoff, bin_size, **kwargs)
        return featurizer

    # @staticmethod
    # def format_fingerprint(fingerprint):
    #     Consider a sigma input that smooths the function
    #     from scipy.ndimage import gaussian_filter1d
    #     new_fp = gaussian_filter1d(fp, sigma=1)



# -*- coding: utf-8 -*-

import importlib.metadata
import logging

from rich.logging import RichHandler

__version__ = importlib.metadata.version("simmate")

# Configure our logger to output timestamps and "SIMMATE" with logs
# Also changes the logging level to info
logging.basicConfig(
    format="%(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        RichHandler(
            show_path=False,
            markup=True,
            rich_tracebacks=True,
        )
    ],
)



# -*- coding: utf-8 -*-

import logging
import shutil
from pathlib import Path

from pymatgen.analysis.transition_state import NEBAnalysis
from pymatgen.io.vasp.outputs import Vasprun as VasprunPymatgen


class Vasprun(VasprunPymatgen):
    @classmethod
    def from_directory(cls, directory: Path = None):

        if not directory:
            directory = Path.cwd()

        vasprun_filename = directory / "vasprun.xml"

        # load the xml file and all of the vasprun data
        try:
            vasprun = cls(
                filename=vasprun_filename,
                exception_on_bad_xml=True,
            )
        except:
            logging.warning(
                "XML is malformed. This typically means there's an error with your"
                " calculation that wasn't caught by your ErrorHandlers. We try"
                " salvaging data here though."
            )
            vasprun = cls(
                filename=vasprun_filename,
                exception_on_bad_xml=False,
            )
            vasprun.final_structure = vasprun.structures[-1]
        # This try/except is just for my really rough calculations
        # where I don't use any ErrorHandlers and still want the final structure
        # regarless of what went wrong. In the future, I should consider writing
        # a separate method for those that loads the CONTCAR and moves on.

        # set source directory for convenience elsewhere
        vasprun.directory = directory

        return vasprun

    @property
    def neb_results(self):

        directory = getattr(self, "directory", None)
        if not directory:
            raise Exception(
                "The Vasprun must have been created with the `from_directory` "
                "method in order to load neb results because it involves loading "
                "more results from files."
            )

        # Make sure there is "*.start" and "*.end" directory present. These
        # will be our start/end folders. We go through all foldernames in the
        # directory and grab the first that matches
        # BUG: For now I assume there are start/end image directories are located
        # in the working directory. These relaxation are actually ran by a
        # separate workflow, which is thus a prerequisite for this workflow.

        # assume foldernames of start and end until proven otherwise
        start_dirname = directory / "start"
        end_dirname = directory / "end"

        for name in directory.iterdir():
            if name.suffix == ".start":
                start_dirname = name
            elif name.suffix == ".end":
                end_dirname = name

        if not start_dirname.exists() or not end_dirname.exists():
            raise Exception(
                "Your NEB calculation finished (possibly successfully). However, "
                "in order to run the workup, Simmate needs the start/end point "
                "relaxations. These should be located in the same directory as "
                "the NEB run and with folder names ending  with '*.start' and"
                " '*.end' (e.g. 'image.start' and image.end' will work)"
            )

        ################
        # BUG: NEBAnalysis.from_dir is broken for all folder structures except
        # when the start/end points are in the 00 and N folders. I therefore
        # need to copy the OUTCAR from the endpoint relaxations to these folders.
        # I don't want to mess with opening a pull request with them / waiting
        # on a new release, so I make this hacky fix here
        new_start_filename = directory / "00" / "OUTCAR"
        # the end filename should be the highest number in the directory
        numbered_dirs = [d for d in directory.iterdir() if d.name.isdigit()]
        numbered_dirs.sort()
        new_end_filename = directory / numbered_dirs[-1] / "OUTCAR"
        # now copy the outcars over
        shutil.copyfile(start_dirname / "OUTCAR", new_start_filename)
        shutil.copyfile(end_dirname / "OUTCAR", new_end_filename)
        ################

        neb_results = NEBAnalysis.from_dir(
            directory,
            # BUG: see bug fix right above this
            # relaxation_dirs=[
            #     "endpoint_relaxation_start",
            #     "endpoint_relaxation_end",
            # ],
        )
        return neb_results



# -*- coding: utf-8 -*-

from django.shortcuts import render

from simmate.database import third_parties
from simmate.database.base_data_types import DatabaseTable
from simmate.database.third_parties import (
    AflowPrototype,
    AflowStructure,
    CodStructure,
    JarvisStructure,
    MatprojStructure,
    OqmdStructure,
)
from simmate.website.core_components.base_api_view import SimmateAPIViewSet


def providers_all(request):
    context = {
        "all_providers": [
            AflowPrototype,
            # AflowStructure,  # Not allowed yet
            CodStructure,
            JarvisStructure,
            MatprojStructure,
            OqmdStructure,
        ]
    }
    template = "third_parties/providers_all.html"
    return render(request, template, context)


class ProviderAPIViewSet(SimmateAPIViewSet):

    template_list = "third_parties/provider.html"
    template_retrieve = "third_parties/entry_detail.html"

    @classmethod
    def get_table(
        cls,
        request,
        provider_name,
        pk=None,
    ) -> DatabaseTable:
        """
        grabs the relevant database table using the URL request
        """
        # using the provider name (which is really just the table name), load
        # the corresponding database table
        provider_table = getattr(third_parties, provider_name)
        return provider_table

    def get_list_context(
        self,
        request,
        provider_name,
    ) -> dict:

        provider_table = getattr(third_parties, provider_name)
        return {"provider": provider_table}

    def get_retrieve_context(
        self,
        request,
        provider_name,
        pk,
    ) -> dict:

        return {
            "provider_name": provider_name,
            "entry_id": pk,
        }



# -*- coding: utf-8 -*-

import logging

from numpy.random import choice, randint
from rich.progress import track

from simmate.toolkit import Composition
from simmate.toolkit.creators.vector import UniformlyDistributedVectors
from simmate.toolkit.symmetry.wyckoff import (
    findValidWyckoffCombos,
    loadAsymmetricUnitData,
    loadWyckoffData,
)


class RandomWySites:
    def __init__(
        self,
        composition: Composition,
        spacegroup_include: list[int] = range(1, 231),
        spacegroup_exclude: list[int] = [],
        coords_generation_method=UniformlyDistributedVectors,
        # note, asymmetric_unit_boundries is included in coords_gen_options
        # automatically so this input is anything extra.
        coords_gen_options: dict = dict(),
        lazily_generate_combinations: bool = True,
    ):

        # make a list of spacegroups that we are allowed to choose from
        self.spacegroup_options = [
            sg for sg in spacegroup_include if sg not in spacegroup_exclude
        ]

        # save the composition for reference
        self.composition = composition

        # the stoichiometry of the composition will be used repeatedly below
        # it's easier to save that here for reference too
        # convert the composition to a stoichiometry list
        # for example, Mg4Si4O12 composition will have stoich=[4,4,12]
        stoichiometry = [int(composition[element]) for element in composition]
        self.stoichiometry = stoichiometry

        # Find valid wyckoff combinatons for all spacegroups and stoichiometry
        # I want the combinations and info to be stored as a dictionary of
        # {spacegroup:combos/info} entries.
        # We split this into 2 dictionaries becuase we reference each a lot and
        # it makes the code below easier.
        self.wy_groupinfo = {}
        self.wy_groupcombos = {}
        # some spacegroups will be incompatible with the given composition, so
        # we keep a list of these
        self.spacegroups_invalid = []

        # Each spacegroup will have a "generator" that manages possible wyckoff
        # combinations and the asymmetric unit associated with the spacegroup.
        # These will be stored as a dictionary of  {spacegroup: generator}
        self.coords_generators = {}
        # Options set in coords_gen_options will apply to all spacegroups, so
        # we store these up front. This because these generators are
        # sometimes accessed lazily
        self.coords_generation_method = coords_generation_method
        self.coords_gen_options = coords_gen_options

        # Generating all wyckoff combinations can take a long time to run in
        # some cases (e.g. high atom counts), so we may want to do it up front.
        # There are also scenarios where we don't want to generate the possible
        # spacegroup combinations until the spacegroup is actually requested.
        # This saves us time when we want just one structure and then are
        # done with this system.
        self.lazily_generate_combinations = lazily_generate_combinations
        if not lazily_generate_combinations:
            logging.info("Generating Wyckoff combinations for every spacegroup")
            for spacegroup in track(self.spacegroup_options):
                self._setup_spacegroup_wyckoff_generator(
                    spacegroup,
                    show_logging=False,
                )
            logging.info("Done.")

        # below, I'll need to repeatedly reference wy_data
        #!!! I want to find a better way of loading wy_data - maybe load
        # only data for spacegroup_options?
        self.wy_data = loadWyckoffData().values

    def new_sites(self, spacegroup=None):

        # parse spacegroup or grab a random one (with necessary lazy-setup)
        spacegroup = self._init_spacegroup(spacegroup)
        # exit if we have an invalid spacegroup
        if not spacegroup:
            return False

        # we need to grab the generators, wyckoff combinations, wyckoff group
        # information for selected spacegroup
        coords_generator = self.coords_generators[spacegroup]
        wy_groupcombos = self.wy_groupcombos[spacegroup]
        wy_groupinfo = self.wy_groupinfo[spacegroup]

        # randomly grab a wy_group combination
        # NOTE: this does not have equal probability of grabbing all wyckoff
        # site combinations because some wy_groups have more wy_site p
        # osibilities than others.
        # choice(sg_combos['ValidCombinations']) doesn't work here, so we grab
        # a random index instead
        random_index = randint(0, len(wy_groupcombos))
        wy_group_combo = wy_groupcombos[random_index]

        # make an empty list to store all the fractional coordinates
        coords_list = []  # regular list.append is faster than numpy.append
        # as well as an empty list for the species
        species_list = []

        # now we need to iterate through each element in stoich, see which
        # wy_group(s) it was assigned to
        for i, wy_groups in enumerate(wy_group_combo):
            # wy_groups represents the combination of wy_group's for a single element
            # len(stoich) == len(wy_group_combo) because indexes are matched
            # up (i.e. stoich[1] was assigned wy_group_combo[1])
            # so we can use the index of wy_groups to grab the element
            element = self.composition.elements[i]
            # wy_groups can be a list of wy_group entries
            for wy_group in wy_groups:
                # now we need to generate the coordinates for this wy_site
                # for that, we need three things:
                # (1) a random wy_site from the assigned wy_group
                # (2) the base wy_site coords corresponding to that wy_site
                #     (i.e. (x,y,z) or (0,y,0))
                # (3) random coords that exist inside the asymmetric unit,
                #     used to eval() the coords with
                # (4) ensure that the result coordinates have not been used already
                # wy_group label is a key to the wy_groupinfo dictionary
                # the wy_sites output is a list of indexes (as numpy array)
                # that correspond to indicies in the loadWyckoffData() output
                wy_sites = wy_groupinfo[wy_group]
                check = False
                while not check:
                    # randomly choose one of these wy_site indexes
                    # I have this inside the while-loop because if a special
                    # wy_site (i.e. (0,0,0)) is picked when it's already in
                    # coords_used, we will get stuck in the while loop
                    wy_site_i = choice(wy_sites)
                    # use the choice to grab the wy_site
                    wy_site = self.wy_data[wy_site_i]
                    # grab the first entry (as all others are symmetrically
                    # equivalent) for coord template
                    wy_coords = wy_site[5]  # index 5 is 'Coordinates'
                    # generate random x,y,z values that are inside the
                    # asymmetric unit
                    x, y, z = coords_generator.new_vector()
                    # place these x,y,z values into the coords where necessary
                    # not sure why I need to use None,dict(x=x,y=y,z=z)
                    # inside eval. Eval() can't use the locals for some reason.
                    final_coords = eval(wy_coords, None, dict(x=x, y=y, z=z))
                    # see if the result is already a coordinate used
                    check = final_coords not in coords_list
                # append the coords and species to our output lists
                coords_list.append(final_coords)
                species_list.append(element)

        # maybe make these into pymatgen Site objects? I don't see any
        # advantage to Site object, because we will later need PeriodicSite
        # objects and there's no direct conversion method
        return species_list, coords_list

    def _init_spacegroup(self, spacegroup):
        # This checks the spacegroup input from a user or grabs a random one.
        # We isolate this into a separate class because of the complex logic
        # required to handle lazy

        # if a spacegroup is not specified, grab a random one from our options
        if not spacegroup:
            # If we generated the combinatons up front, we can just grab a spacegroup
            # from all the valid options.
            if not self.lazily_generate_combinations:
                spacegroup = choice(self.spacegroup_options)

            # Otherwise, we need to randomly select a symmetry system until we
            # find a valid one
            else:
                spacegroup = -1  # gives non-existing spacegroup to start the loop

                # check if the spacegroup is valid & analyzed before
                while not spacegroup in self.coords_generators:

                    # bug-note: this loop should not go on endlessly, because
                    # spacegroup=1 should **always** be valid and it will be
                    # selected eventually in extreme cases.

                    # grab a new spacegroup
                    spacegroup = choice(self.spacegroup_options)

                    # generate the possible combinations.
                    self._setup_spacegroup_wyckoff_generator(spacegroup)

        # if the user provided a spacegroup, we still need to check if its been
        # analyzed and is valid.
        else:

            # make sure wyckoff combos have been analyzed before
            if (
                self.lazily_generate_combinations
                and spacegroup not in self.coords_generators
                and spacegroup not in self.spacegroups_invalid
            ):
                # generate the possible combinations.
                self._setup_spacegroup_wyckoff_generator(spacegroup)

            # Make sure the spacegroup is valid for the given composition
            if spacegroup in self.spacegroups_invalid:
                logging.warning(
                    f"Spacegroup {spacegroup} is invalid for {self.composition}"
                )
                return False

        return spacegroup

    def _setup_spacegroup_wyckoff_generator(
        self, spacegroup: int, show_logging: bool = True
    ):
        if show_logging:
            logging.info(
                f"Generating possible wyckoff combinations for spacegroup {spacegroup}"
            )
        sg_combo_data = findValidWyckoffCombos(self.stoichiometry, spacegroup)
        if show_logging:
            logging.info("Done generating combinations.")
        # If the spacegroup + stoich combination has no valid combinations,
        # the generator will not work
        if not sg_combo_data["ValidCombinations"]:
            self.spacegroups_invalid.append(spacegroup)
            # we also need to update all the spacegroup_options now that we have
            # a new invalid one.
            self.spacegroup_options = [
                sg
                for sg in self.spacegroup_options
                if sg not in self.spacegroups_invalid
            ]
            # exit the function -- false indicates we have an invalid spacegroup
            return False
        # Otherwise we have valid combos that we want to store -- so that
        # we don't need to calculate them again.
        self.wy_groupinfo.update({spacegroup: sg_combo_data["WyckoffGroups"]})
        self.wy_groupcombos.update({spacegroup: sg_combo_data["ValidCombinations"]})

        # Next we need to setup the asymmetric unit boundry conditions.

        # grab a copy of the user inputs for reference
        # we will manipulate final_options for each spacegroup for use
        # in the generator
        final_options = self.coords_gen_options.copy()

        # establish the coords generator with asymmetric_unit_boundries
        # this process is a little more complicated because we require
        # that the asymmetric_unit_boundries be included in extra_conditions
        asym_bounds = asymmetric_unit_boundries(spacegroup)

        # if any extra_conditions were provided by the user, we want to
        # combine them with the asym conditions
        if "extra_conditions" in self.coords_gen_options.keys():
            all_bounds = self.coords_gen_options["extra_conditions"] + asym_bounds
            final_options.update({"extra_conditions": all_bounds})
        # otherwise, the asym_bounds are the only boundry conditions for
        # the coordinates
        else:
            final_options.update({"extra_conditions": asym_bounds})

        # now that we have the coords_gen_options updated, we can make the generator
        coords_generator = self.coords_generation_method(**final_options)

        # add store this generator
        self.coords_generators.update({spacegroup: coords_generator})

        # exit the function -- true indicates we have a valid spacegroup
        return True


# Grab the boundry conditions for the asymmetric unit of a spacegroup's unitcell
def asymmetric_unit_boundries(spacegroup, asym_data=loadAsymmetricUnitData()):

    # spacegroup = spacegroup that we are interested in
    # asym_data = result of loadAsmmetricUnitData() - this is in the header for speed reasons

    # grab the asymmetric unit conditons from the asym_data
    # the array indicies (0-229) correspond to all spacegroups (1-230) respectively.
    # So if I want data for spacegroup 230, I need index 229 (i.e. spacegroup-1)
    asym_bounds = asym_data[spacegroup - 1]

    # convert the string of asym_data into a list of boundry conditions that we can test
    # so here we change things to something python can read and iterate through
    asym_bounds = asym_bounds.replace("≤", "<=")  # convert symbols
    # Is there a way I can do these next three lines in a single command...?
    # or maybe only run them for the spacegroups that need them?
    asym_bounds = asym_bounds.replace("2y", "(2*y)")  # convert
    asym_bounds = asym_bounds.replace("2x", "(2*x)")  # convert
    asym_bounds = asym_bounds.replace("-4x", "(-4*y)")  # convert
    asym_bounds = asym_bounds.split(";")  # convert str into a list

    return asym_bounds



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.electronic_structure.base_full import (
    ElectronicStructureWorkflow,
)
from simmate.calculators.vasp.workflows.electronic_structure.matproj_band_structure_hse import (
    ElectronicStructure__Vasp__MatprojBandStructureHse,
)
from simmate.calculators.vasp.workflows.electronic_structure.matproj_density_of_states_hse import (
    ElectronicStructure__Vasp__MatprojDensityOfStatesHse,
)
from simmate.calculators.vasp.workflows.static_energy.matproj_hse import (
    StaticEnergy__Vasp__MatprojHse,
)


class ElectronicStructure__Vasp__MatprojHseFull(ElectronicStructureWorkflow):
    """
    runs DOS and BS at Materials Project settings
    """

    description_doc_short = "runs DOS and BS at Materials Project settings"

    # table not implemented yet. This is a placeholder
    database_table = ElectronicStructure__Vasp__MatprojBandStructureHse.database_table

    static_energy_workflow = StaticEnergy__Vasp__MatprojHse
    band_structure_workflow = ElectronicStructure__Vasp__MatprojDensityOfStatesHse
    density_of_states_workflow = ElectronicStructure__Vasp__MatprojBandStructureHse



# -*- coding: utf-8 -*-

import pytest

from simmate.calculators.vasp.workflows.electronic_structure.matproj_density_of_states import (
    ElectronicStructure__Vasp__MatprojDensityOfStates,
)
from simmate.conftest import SimmateMockHelper, copy_test_files


@pytest.mark.django_db
def test_density_of_states_run(sample_structures, tmp_path, mocker):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="band_structure.zip",
    )

    SimmateMockHelper.mock_vasp(mocker)

    # run the full workflow, where the output files were pre-generated with
    # a specific structures
    structure = sample_structures["Fe_mp-13_primitive"]
    ElectronicStructure__Vasp__MatprojDensityOfStates.run(
        structure=structure,
        directory=tmp_path,
    )

    # check output files
    summary_filename = tmp_path / "simmate_summary.yaml"
    plot_filename = tmp_path / "dos_diagram.png"
    assert summary_filename.exists()
    assert plot_filename.exists()



# -*- coding: utf-8 -*-

"""
For adding new selection methods, DEAP is a good reference:
  https://github.com/DEAP/deap/blob/master/deap/tools/selection.py
  https://deap.readthedocs.io/en/master/api/tools.html

Their selection methods take a list of individuals that each have a .fitness
attribute. Instead, I pass individuals here as a pandas dataframe and have
them specify a "fitness_column"

Another good reference is Wikipedia
  https://en.wikipedia.org/wiki/Selection_(genetic_algorithm)

NOTE: I assume lower fitness is better! This may change in the future because
other selection methods throw issues with lower=better and even more issues
when we have negative values
"""


class Selector:
    def select(self, nselect, individuals, fitness_column):
        # The select method should take in two arguments:
        # individuals = a pandas dataframe with columns "id" and "fitness"
        # n = the number of individuals that we want to choose from those fitnesses
        # The select method should then return a list/tuple of the indicies
        # of the fitnesses that were selected
        raise NotImplementedError()

    @classmethod
    @property
    def name(cls):
        """
        A nice string name for the selector. By default it just returns the name
        of this class.
        """
        return cls.__name__

    def select_from_datatable(
        self,
        nselect,
        datatable,
        ranking_column: str = None,
        query_limit: str = None,
    ):

        # our selectors just require a dataframe where we specify the fitness
        # column. So we query our individuals database to give this as an input.
        datatable_cleaned = datatable
        if ranking_column:
            datatable_cleaned = datatable_cleaned.order_by(ranking_column)
        if query_limit:
            datatable_cleaned = datatable_cleaned[:query_limit]
        individuals_df = datatable_cleaned.to_dataframe()
        # NOTE: I assume we'll never need more than the best 200 structures, which
        # may not be true in special cases.

        # From these individuals, select our parent structures
        parents_df = self.select(nselect, individuals_df, ranking_column)

        # grab the id column of the parents and convert it to a list
        parent_ids = parents_df.id.values.tolist()

        # Now lets grab these structures from our database and convert them
        # to a list of pymatgen structures.
        # We have to make separate queries for this instead of doing "id__in=parent_ids".
        # This is because (1) order may be important and (2) we may have duplicate
        # entries. For example, a hereditary mutation can request parent ids of [123,123]
        # in which case we want to give the same input structure twice!
        parent_structures = [
            datatable.only("structure").get(id=parent_id).to_toolkit()
            for parent_id in parent_ids
        ]

        # When there's only one structure selected we return the structure and
        # id independently -- not within a list
        if nselect == 1:
            parent_ids = parent_ids[0]
            parent_structures = parent_structures[0]

        # for record keeping, we also want to return the ids for each structure
        return parent_ids, parent_structures


class TruncatedSelection(Selector):
    """
    Truncated selection limits the parent selection to the top X% of individuals,
    and then within this top section, every individual has an equal chance of
    being selected.
    """

    def __init__(self, percentile=0.05, ntruncate_min=5, allow_duplicate=True):

        # aim to grab this percent of individuals
        self.percentile = percentile

        # at minimum, make sure we can select from this number of best individuals
        self.ntruncate_min = ntruncate_min

        # whether we can select the same individual more than once
        self.allow_duplicate = allow_duplicate

    def select(self, nselect, individuals, fitness_column):

        # truncate the population to those with energies in the lowest X%. This
        # value should be greater or equal to our minimum set above
        ntruncate = int(len(individuals) * self.percentile)
        if ntruncate < self.ntruncate_min:
            ntruncate = self.ntruncate_min

        # limit our potential selection to the top X% and then randomly select
        # the correct number of parents
        df_truncated = individuals.nsmallest(ntruncate, fitness_column)
        df_parents = df_truncated.sample(nselect, replace=self.allow_duplicate)

        # return the list of indexes to be selected
        return df_parents



# -*- coding: utf-8 -*-

from .diffusion import (
    Diffusion__Vasp__NebAllPathsMit,
    Diffusion__Vasp__NebFromEndpointsMit,
    Diffusion__Vasp__NebFromImagesMit,
    Diffusion__Vasp__NebFromImagesMvlCi,
    Diffusion__Vasp__NebSinglePathMit,
)
from .dynamics import (
    Dynamics__Vasp__Matproj,
    Dynamics__Vasp__Mit,
    Dynamics__Vasp__MvlNpt,
)
from .electronic_structure import (
    ElectronicStructure__Vasp__MatprojBandStructure,
    ElectronicStructure__Vasp__MatprojBandStructureHse,
    ElectronicStructure__Vasp__MatprojDensityOfStates,
    ElectronicStructure__Vasp__MatprojDensityOfStatesHse,
    ElectronicStructure__Vasp__MatprojFull,
    ElectronicStructure__Vasp__MatprojHseFull,
)
from .population_analysis import (
    PopulationAnalysis__Vasp__ElfMatproj,
    PopulationAnalysis__VaspBader__BadelfMatproj,
    PopulationAnalysis__VaspBader__BaderMatproj,
    StaticEnergy__Vasp__PrebadelfMatproj,
    StaticEnergy__Vasp__PrebaderMatproj,
)
from .relaxation import (
    Relaxation__Vasp__Matproj,
    Relaxation__Vasp__MatprojHse,
    Relaxation__Vasp__MatprojMetal,
    Relaxation__Vasp__MatprojScan,
    Relaxation__Vasp__Mit,
    Relaxation__Vasp__MvlGrainboundary,
    Relaxation__Vasp__MvlNebEndpoint,
    Relaxation__Vasp__MvlSlab,
    Relaxation__Vasp__Quality00,
    Relaxation__Vasp__Quality01,
    Relaxation__Vasp__Quality02,
    Relaxation__Vasp__Quality03,
    Relaxation__Vasp__Quality04,
    Relaxation__Vasp__Staged,
)
from .static_energy import (
    StaticEnergy__Vasp__Matproj,
    StaticEnergy__Vasp__MatprojHse,
    StaticEnergy__Vasp__MatprojScan,
    StaticEnergy__Vasp__Mit,
    StaticEnergy__Vasp__MvlNebEndpoint,
    StaticEnergy__Vasp__Quality04,
)



# -*- coding: utf-8 -*-

import inspect
import json
import logging
import re
import uuid
from pathlib import Path

import cloudpickle
import toml
import yaml

import simmate
from simmate.database.base_data_types import Calculation
from simmate.utilities import copy_directory, get_directory, make_archive
from simmate.workflow_engine.execution import SimmateExecutor, WorkItem


class DummyState:
    """
    This class is meant to emulate Prefect States. By wrapping a result into
    State, we enable higher-level features that depend on a call to
    `state.result()`.

    This class should not be used directly as it is automatically applied with
    the `Workflow.run` method
    """

    def __init__(self, result):
        self._result = result

    def result(self):
        return self._result

    @staticmethod
    def is_completed():
        return True


class Workflow:
    """
    An abstract base class for all Simmate workflows.
    """

    version: str = simmate.__version__
    """
    Version number for this flow. Defaults to the Simmate version 
    (e.g. "0.7.0").
    """

    description_doc_short: str = None
    """
    A quick description for this workflow. This will be shown in the website UI
    in the list-view of all different workflow presets.
    """

    use_database: bool = True
    """
    Whether to use Simmate database features or not.
    
    This includes calling the `_register_calculation` and 
    `_update_database_with_results` methods attached to this workflow.
    
    `_register_calculation` will save a database entry before the workflow
    starts. This is useful to keep track of workflows that have been
    submitted/started but haven't finished yet.
    
    `_update_database_with_results` saves the output of the `workup` method
    to the database entry -- and this the same entry that was created by 
    `_register_calculation`.
    """

    _parameter_methods = ["run_config", "_run_full"]
    """
    List of methods that allow unique input parameters. This helps track where
    `**kwargs` are passed and let's us gather the inputs in one place.
    """

    # -------------------------------------------------------------------------
    # Core methods that handle how and what a workflow run does
    # and how it is submitted
    # -------------------------------------------------------------------------

    @classmethod
    def run(cls, **kwargs) -> DummyState:
        """
        The highest-level method to run a workflow. This includes calling
        `run_config` and all of the extra steps involved (such as registering
        a calculation to the database and creating the working directory).

        For advanced users, we recommend reading through the `_run_full` method
        to see the full workflow logic.
        """
        # Note: this is a separate method and wrapper around run_full because
        # we want to allow Prefect executor to overwrite this method.
        result = cls._run_full(**kwargs)
        state = DummyState(result)
        return state

    @classmethod
    def _run_full(
        cls,
        run_id: str = None,
        directory: Path | str = None,
        compress_output: bool = False,
        source: dict = None,
        **kwargs,
    ):
        """

        The full logic for a workflow run. This deserialize input parameters,
        register the calculation to the database, call the defined
        `run_config` command, and save results to the database.

        This method should not be called directly. Use the `run` method instead,
        which will properly convert to a `Prefect` workflow (if necessary) and
        return a State-like object.

        #### Parameters

        - `run_id`:
            A unique id that is assigned to the workflow run. If an ID was not
            provided, a randomly-generated id from
            [UUID4](https://docs.python.org/3/library/uuid.html#uuid.uuid4)
            is used. We recommend leaving this at default.

         - `directory`:
             The directory to run everything in. This is passed to the ulitities
             function simmate.ulitities.get_directory, which generates a
             unique foldername if not provided. This will be converted into
             a `pathlib.Path` object.

        - `compress_output`:
            Whether to compress the directory to a zip file at the end of the
            task run. After compression, it will also delete the directory.
            The default is False.

        - `source`: (EXPERIMENTAL FEATURE)
            The source is JSON-like information that indicates where the
            input for this calculation came from, which could be a number
            of things including (another calculation, a simple comment, etc.).
            This is useful if you want to label results in your database.

        """
        # This method is isolated only because we want to wrap it as a prefect
        # workflow in some cases.
        logging.info(f"Starting '{cls.name_full}'")
        kwargs_cleaned = cls._load_input_and_register(
            run_id=run_id,
            directory=directory,
            compress_output=compress_output,
            source=source,
            **kwargs,
        )

        # Finally run the core part of the workflow. This should return a
        # dictionary object if we have "use_database=True", but can be
        # any python object if "use_database=False"
        results = cls.run_config(**kwargs_cleaned)

        # save the result to the database
        if cls.use_database:

            # make sure the workflow is returning a dictionary that be used
            # to update the database columns. None is also allowed as it
            # represents an empty dictionary
            if not isinstance(results, dict) and results != None:
                raise Exception(
                    "When using a database table, your `run_config` method must "
                    "return a dictionary object. The dictionary is used to "
                    "update columns in your table entry and is therefore a "
                    "required format. If you do not want to save to the database "
                    "(and avoid this message), set `use_database=False`"
                )
            logging.info("Saving to database and writing outputs")
            database_entry = cls._update_database_with_results(
                results=results if results != None else {},
                directory=kwargs_cleaned["directory"],
                run_id=kwargs_cleaned["run_id"],
            )

        # if requested, compresses the directory to a zip file and then removes
        # the directory.
        if compress_output:
            logging.info("Compressing result to a ZIP file.")
            make_archive(kwargs_cleaned["directory"])

        # If we made it this far, we successfully completed the workflow run
        logging.info(f"Completed '{cls.name_full}'")

        # If we are using the database, then we return the database object.
        # Otherwise, we want to return the original result from run_config
        return database_entry if cls.use_database else results

    @classmethod
    def run_cloud(
        cls,
        tags: list[str] = None,
        **kwargs,
    ):
        """
        Submits the workflow run to cloud database to be ran by a worker.

        #### Parameters

        - `return_state`:
            Whether to return a State-like object or just the run_id. Defaults
            to true, which returns a State.

        - `tags`:
            A list of flags/labels/tags that the workflow run should be scheduled
            with. This helps with limiting which workers are allow to pickup
            and run the workflow. Defaults to the `tags` property of the
            workflow.
        """

        logging.info(f"Submitting new run of `{cls.name_full}` to cloud")

        # To help with tracking the flow in cloud, we load all of the inputs up
        # front. This will include creating a run_id for us.
        #
        # If we are submitting using a filename, we don't want to
        # submit to a cluster and have the job fail because it doesn't have
        # access to the file. We therefore go through the full load_input process
        kwargs_cleaned = cls._load_input_and_register(
            setup_directory=False,
            write_metadata=False,
            **kwargs,
        )

        # Some backends can't pickle input parameters, so we need to serialize
        # them before submission to the queue.
        parameters_serialized = cls._serialize_parameters(**kwargs_cleaned)

        state = SimmateExecutor.submit(
            cls._run_full,  # should this be the run method...?
            tags=tags or cls.tags,
            **parameters_serialized,
        )

        logging.info(f"Successfully submitted (workitem_id={state.pk})")

        return state

    @classmethod
    def run_config(cls, **kwargs) -> any:
        """
        The workflow method, which can be overwritten when inheriting from this
        class. This can be either a `staticmethod` or `classmethod`.
        """
        raise NotImplementedError(
            "When creating a custom workflow, make sure you set a run_config method!"
        )

    # -------------------------------------------------------------------------
    # Methods for running/submitting workflows from input files (e.g. YAML/TOML)
    # -------------------------------------------------------------------------

    @classmethod
    def run_from_file(cls, filename: Path) -> DummyState:
        """
        Runs a workflow locally where parameters are loaded from a yaml or toml file
        """
        workflow, parameters = cls._load_settings_file(filename)
        state = workflow.run(**parameters)
        return state

    @classmethod
    def run_cloud_from_file(cls, filename: Path):
        """
        Submits a workflow to cloud for remote running where parameters are loaded
        from a yaml or toml file
        """
        workflow, parameters = cls._load_settings_file(filename)
        state = workflow.run_cloud(**parameters)
        return state

    @classmethod
    def _load_settings_file(cls, filename: Path):  # -> tuple[Workflow, dict]

        filename = Path(filename)

        # Load the settings to a dictionary using whichever format is given
        with filename.open() as file:
            if filename.suffix == ".yaml":
                parameters = yaml.full_load(file)
            elif filename.suffix == ".toml":
                parameters = toml.load(file)
            # elif filename.suffix == "json":  TODO consider allowing json
            #     parameters = json.load(file)
            else:
                raise Exception(
                    "Unknown input file format provided. Please make sure your "
                    "filename ends in either `.yaml` or `.toml`"
                )

        # If a workflow name is given, then we are actually calling this method
        # from the base workflow class -- and need to grab the proper workflow
        # object.
        if "workflow_name" in parameters.keys():

            # we are loading an external workflow, which will depend on this base
            # class. We need a local import to prevent circular-import bug.
            from simmate.workflows.utilities import get_workflow

            workflow = get_workflow(
                # we pop the workflow name so that it is also removed from the
                # rest of kwargs
                workflow_name=parameters.pop("workflow_name"),
            )
            return workflow, parameters

        # Otherwise we should be calling the proper workflow subclass already
        # and just return the workflow class itself.
        else:
            return cls, parameters

    # -------------------------------------------------------------------------
    # Methods that interact with the Executor class in order to see what
    # has been submitted to cloud.
    # -------------------------------------------------------------------------

    @classmethod
    @property
    def nflows_submitted(cls) -> int:
        """
        Queries the Simmate database to see how many workflows are in a
        running or pending state.
        """
        return WorkItem.objects.filter(
            status__in=["P", "R"],
            tags=cls.tags,
        ).count()

    # -------------------------------------------------------------------------
    # Methods that help with accessing the database and saving results
    # -------------------------------------------------------------------------

    @classmethod
    @property
    def database_table(cls) -> Calculation:
        """
        The database table where calculation information (such as the run_id)
        is stored. The table should use `simmate.database.base_data_types.Calculation`
        as one of its mix-ins.
        """
        # OPTIMIZE: a mapping dictionary or some standardized way to name
        # database tables would simplify this.

        flow_type = cls.name_type
        flow_preset = cls.name_preset

        if flow_type == "relaxation":
            from simmate.database.base_data_types import Relaxation

            return Relaxation
        elif flow_type == "static-energy":
            from simmate.database.base_data_types import StaticEnergy

            return StaticEnergy
        elif flow_type == "electronic-structure":
            if "band-structure" in flow_preset:
                from simmate.database.base_data_types import BandStructureCalc

                return BandStructureCalc
            elif "density-of-states" in flow_preset:
                from simmate.database.base_data_types import DensityofStatesCalc

                return DensityofStatesCalc
        elif flow_type == "population-analysis":
            from simmate.database.base_data_types import PopulationAnalysis

            return PopulationAnalysis
        elif flow_type == "dynamics":
            from simmate.database.base_data_types import Dynamics

            return Dynamics
        elif flow_type == "diffusion":
            if "from-images" in flow_preset or "single-path" in flow_preset:
                from simmate.database.base_data_types import MigrationHop

                return MigrationHop
            elif "all-paths" in flow_preset:
                from simmate.database.base_data_types import DiffusionAnalysis

                return DiffusionAnalysis

        raise NotImplementedError(
            "Unable to detect proper database table. Are you sure your workflow "
            "should be using a table? If not, set `use_database=False` on your "
            "workflow as shown in the 'basic' example workflow from the guides."
        )

    @classmethod
    @property
    def all_results(cls):  # -> SearchResults
        """
        Filters results from the database table down to the results from this
        workflow (i.e. matching workflow_name)
        """
        return cls.database_table.objects.filter(workflow_name=cls.name_full).all()

    @classmethod
    def _update_database_with_results(
        cls,
        results: dict,
        run_id: str,
        directory: Path,
    ) -> Calculation:
        """
        Take the output of the `run_config` and any extra information and
        saves it to the database.

        An output summary is also written to file for quick viewing.
        """

        # load the calculation entry for this workflow run. This should already
        # exist thanks to the load_input_and_register task.
        calculation = cls.database_table.from_run_context(
            run_id=run_id,
            workflow_name=cls.name_full,
            workflow_version=cls.version,
        )

        # Now update the calculation entry with our results. Typically, all of this
        # is handled by the calculation table's "update_from" methods, but in
        # rare cares, we may want to attach an update method directly to the
        # workflow class. I can only imagine this is used when...
        #   (1) workflow attributes are important during the update
        #   (2) when several workflows share a table and need to isolate
        #       their workup method (e.g. the MigrationHop table for NEB)
        if hasattr(cls, "update_database_from_results"):
            # The attribute can also be set to false to disable updates
            if cls.update_database_from_results:
                cls.update_database_from_results(
                    calculation=calculation,
                    results=results,
                    directory=directory,
                )
        # Otherwise we hand this off to the database object
        else:
            calculation.update_from_results(
                results=results,
                directory=directory,
            )

        # write the output summary to file
        calculation.write_output_summary(directory)
        # TODO: consider making this optional to improve speedup

        return calculation

    @classmethod
    def load_completed_calc(cls, directory: Path):
        # TODO: maybe load the yaml file to get extra kwargs, run_id, etc.
        return cls.database_table.from_directory(directory)

    # -------------------------------------------------------------------------
    # Properties that enforce the naming convention for workflows
    # -------------------------------------------------------------------------

    @classmethod
    @property
    def name_full(cls) -> str:
        """
        Standardized name of the workflow. This converts the class name like so:
        `Static_Energy__VASP__Matproj` --> `static-energy.vasp.matproj`
        """
        if not len(cls.__name__.split("__")) == 3:
            raise Exception("Make sure you are following Simmate naming conventions!")

        # convert to dot format
        name = cls.__name__.replace("__", ".")

        # adds a hyphen between each capital letter
        # copied from https://stackoverflow.com/questions/199059/
        name = re.sub(r"(\w)([A-Z])", r"\1-\2", name)

        return name.lower()

    @classmethod
    @property
    def name_type(cls) -> str:
        """
        Name of the Project this workflow is associated with. This is the first
        portion of the flow name (e.g. "static-energy")
        """
        return cls.name_full.split(".")[0]

    @classmethod
    @property
    def name_calculator(cls) -> str:
        """
        Name of the calculator this workflow is associated with. This is the second
        portion of the flow name (e.g. "vasp")
        """
        return cls.name_full.split(".")[1]

    @classmethod
    @property
    def name_preset(cls) -> str:
        """
        Name of the settings/preset this workflow is associated with. This is the third
        portion of the flow name (e.g. "matproj" or "matproj-prebader")
        """
        return cls.name_full.split(".")[2]

    @classmethod
    @property
    def tags(cls) -> list[str]:
        """
        Lists of tags to submit a the workflow with when using run_cloud.
        """
        return [
            "simmate",
            cls.name_type,
            cls.name_calculator,
            cls.name_full,
        ]

    # -------------------------------------------------------------------------
    # Properties/method that set website UI documentation and help users
    # explore input options and settings used.
    # -------------------------------------------------------------------------

    # BUG: naming this `description` causes issues.
    # See https://github.com/PrefectHQ/prefect/issues/3911
    @classmethod
    @property
    def description_doc(cls) -> str:
        """
        This simply returns the documentation string of this workflow -- so this
        is the same as `__doc__`. This attribute is only defined for beginners
        to python and for use in django templates for the website interface.
        """
        return cls.__doc__ or "No description provided"

    @classmethod
    @property
    def parameter_names(cls) -> list[str]:
        """
        Gives a list of all the parameter names for this workflow.
        """
        # Iterate through and grab the parameters for the core methods. We also
        # sort parameters alphabetically for consistent results.
        parameters = []
        for method in cls._parameter_methods:
            sig = inspect.signature(getattr(cls, method))
            for parameter in list(sig.parameters):
                if parameter not in parameters and parameter != "kwargs":
                    parameters += [parameter]
        parameters.sort()
        return parameters

    @classmethod
    @property
    def parameter_names_required(cls) -> list[str]:
        """
        Gives a list of all the required parameter names for this workflow.
        """
        parameters = []
        for method in cls._parameter_methods:
            sig = inspect.signature(getattr(cls, method))
            for parameter, val in sig.parameters.items():
                if (
                    parameter not in parameters
                    and parameter not in ["directory", "kwargs"]
                    and val.default == val.empty
                ):
                    parameters += [parameter]
        parameters.sort()
        return parameters

    @classmethod
    def show_parameters(cls):
        """
        Prints a list of all the parameter names for this workflow.
        """
        # use yaml to make the printout pretty (no quotes and separate lines)
        print("REQUIRED PARAMETERS")
        print("--------------------")
        print(yaml.dump(cls.parameter_names_required))

        print("OPTIONAL PARAMETERS (+ their defaults):")
        print("---------------------------------------")
        as_list = [{k: v} for k, v in cls.parameter_defaults.items()]
        print(yaml.dump(as_list))
        print("*** 'null' indicates the parameter is set with advanced logic\n")

    @classmethod
    @property
    def parameter_defaults(cls):
        """
        Inspect the run_config and other methods to see what the default
        value for an input parameter is.
        """
        defaults = {}
        for method in cls._parameter_methods:
            sig = inspect.signature(getattr(cls, method))
            for parameter_name in sig.parameters.keys():
                if parameter_name in cls.parameter_names:
                    value = sig.parameters[parameter_name]
                    if value.default == value.empty:
                        continue  # dont save if the value is none
                    defaults.update({parameter_name: value.default})

        # For a series of parameters, we want their default values to be loaded
        # from class attributes if they are not set. To do this, we first check
        # if the parameter is set in our kwargs dictionary and check if the
        # value is set to "None". If it is, then we change to the value set as
        # the class attribute.
        for parameter_name in cls.parameter_names:
            if hasattr(cls, parameter_name):
                defaults[parameter_name] = getattr(cls, parameter_name)

        return defaults

    @classmethod
    def get_config(cls):
        """
        Grabs the overall settings from the class in order to let users briefly
        look at key parameters.

        Using this method is scripts is not recommended.
        """
        return dict(
            use_database=cls.use_database,
            name_full=cls.name_full,
            parameter_names=cls.parameter_names,
            parameter_names_required=cls.parameter_names_required,
        )

    @classmethod
    def show_config(cls):
        """
        Takes the result of get_config and prints it in a yaml format that is
        easier to read.
        """
        config = cls.get_config()
        print(yaml.dump(config))

    # -------------------------------------------------------------------------
    # Properties/method that configure registration of workflow runs and the
    # parameters each is called with.
    # -------------------------------------------------------------------------

    @classmethod
    def _load_input_and_register(
        cls,
        setup_directory: bool = True,
        write_metadata: bool = True,
        **parameters: any,
    ) -> dict:
        """
        How the input was submitted as a parameter depends on if we are submitting
        to Prefect Cloud, running the flow locally, or even continuing from a
        previous calculation.  Here, we use a task to convert the input to a toolkit
        object and (if requested) provide the directory as well.

        For example, a common input parameter for workflows is "structure", which
        can be provided a number of ways:
            - a filename
            - a json string
            - a dictionary pointing to a database entry
            - a toolkit Structure object
            - etc...
        Even though all of these inputs are accepted, `workflow.run` always expects
        python objects, so this utility converts the input to a toolkit Structure
        object.

        `register_run` allows us to skip the database step if the database_table
        isn't properly set yet. This input is a temporary fix for the
        diffusion/from-images workflow.

        `copy_previous_directory` is only used when we are pulling a structure from a
        previous calculation. If copy_previous_directory=True, then the directory
        parameter is ignored.

        `**parameters` includes all parameters and anything extra that you want saved
        to simmate_metadata.yaml
        """

        # OPTIMIZE: consider splitting this task into load_structure, load_directory,
        # and register_calc so that our flow_visualize looks cleaner

        # OPTIMIZE: Customized workflows cause a lot of special handling in this task
        # so it may be worth isolating these into a separate task.

        # ---------------------------------------------------------------------

        # STEP 1: clean parameters

        parameters_cleaned = cls._deserialize_parameters(**parameters)

        # ---------------------------------------------------------------------

        # STEP 2: Load the source of the input object

        source = parameters.get("source", None)

        # User-given source takes the top priority
        if source:
            source_cleaned = source
        # Check if we have a primary input loaded from a past calculation and
        # default to that as the source.
        else:
            # If no source was given, we try to dynamically determine the source
            # using the "primary" input.
            # OPTIMIZE: Is there a better way to do this?

            # Currently I just set a priority of possible parameters that can be
            # the primary input. I go through each one at a time until I find one
            # that was provided -- then I exit with that parameter's value.
            primary_input = None
            for primary_input_key in [
                "structure",
                "migration_hop",
                "supercell_start",
            ]:
                if primary_input_key in parameters_cleaned.keys():
                    # note we grab the deserialized input
                    primary_input = parameters_cleaned.get(primary_input_key, None)
                    break

            if (
                primary_input
                and hasattr(primary_input, "source")
                and primary_input.source
            ):
                source_cleaned = primary_input.source

            else:
                source_cleaned = None

        # before setting the source, we also need it to be json-seralized
        # for the database
        source_cleaned = cls._serialize_parameters(source=source_cleaned)["source"]

        # SPECIAL CASE for customized flows
        if "workflow_base" not in parameters_cleaned:
            parameters_cleaned["source"] = source_cleaned
        else:
            parameters_cleaned["input_parameters"]["source"] = source_cleaned

        # ---------------------------------------------------------------------

        # STEP 3: Load the directory (and copy over from an old directory if necessary)

        if setup_directory:
            # Start by creating a new directory or grabbing the one given. We create
            # this directory immediately (rather than just passing the name to the
            # S3Task). We do this because NestedWorkflows often use a parent directory
            # to organize results.
            directory = parameters.get("directory", None)
            directory_cleaned = get_directory(directory)

            # if the user requested, we grab the previous directory as well
            copy_previous_directory = parameters.get("copy_previous_directory", None)
            if copy_previous_directory:

                # BUG: I should switch this to checking the source input arg

                if not primary_input:
                    raise Exception(
                        "No primary input detected, which is required for copying "
                        "past directories. This is an experimental feature so "
                        "please contact our team for more help."
                    )

                # catch incorrect use of this function
                if not primary_input.database_object:
                    raise Exception(
                        "There isn't a previous directory available! Your source "
                        "structure must point to a past calculation to use this feature."
                    )

                # the past directory should be stored on the input object
                previous_directory = Path(primary_input.database_object.directory)

                # Copy over all files except simmate ones (we have no need for the
                # summaries or error archives)
                copy_directory(
                    directory_old=previous_directory,
                    directory_new=directory_cleaned,
                    ignore_simmate_files=True,
                )

            # SPECIAL CASE for customized flows
            if "workflow_base" not in parameters_cleaned:
                parameters_cleaned["directory"] = directory_cleaned
            else:
                parameters_cleaned["input_parameters"]["directory"] = directory_cleaned

        # ---------------------------------------------------------------------

        # STEP 4: Register the calculation so the user can follow along in the UI
        # and also see which structures/runs have been submitted aready.

        parameters_cleaned["run_id"] = (
            parameters_cleaned.get("run_id", None) or cls._get_new_run_id()
        )

        if cls.use_database:
            cls._register_calculation(**parameters_cleaned)

        # ---------------------------------------------------------------------

        # STEP 5: Write metadata file for user reference

        if write_metadata:
            # convert back to json format. We convert back rather than use the original
            # to ensure the input data is all present. For example, we want to store
            # structure data instead of a filename in the metadata.
            # SPECIAL CASE: "if ..." used to catch customized workflows
            parameters_serialized = (
                cls._serialize_parameters(**parameters_cleaned)
                if "workflow_base" not in parameters_cleaned
                else parameters
            )

            # We want to write a file summarizing the inputs used for this
            # workflow run. This allows future users to reproduce the results if
            # desired -- and it also allows us to load old results into a database.
            input_summary = dict(
                _WORKFLOW_NAME_=cls.name_full,
                _WORKFLOW_VERSION_=cls.version,
                **parameters_serialized,
            )

            # now write the summary to file in the same directory as the calc.
            # check the directory and see how many other "simmate_metadata*.yaml" files
            # already exist. Our new file will be based off of this. Simply,
            # if the biggest number is 04 --> then we will write a file named
            # simmate_metadata__05.yaml to keep the count going.
            count = (
                len(
                    [
                        f
                        for f in directory_cleaned.iterdir()
                        if f.name.startswith("simmate_metadata_")
                    ]
                )
                + 1
            )
            count_str = str(count).zfill(2)
            input_summary_file = (
                directory_cleaned / f"simmate_metadata_{count_str}.yaml"
            )
            with input_summary_file.open("w") as file:
                content = yaml.dump(input_summary)
                file.write(content)

        # ---------------------------------------------------------------------

        # Finally we just want to return the dictionary of cleaned parameters
        # to be used by the workflow
        return parameters_cleaned

    @staticmethod
    def _get_new_run_id():
        """
        Generates a random id to use as a workflow run id.

        This is called automatically within `_load_input_and_register`
        """
        # This is a separate method in order to allow the prefect executor to
        # overwrite this method.
        unique_id = str(uuid.uuid4())
        return unique_id

    @classmethod
    @property
    def _parameters_to_register(cls) -> list[str]:
        """
        A list of input parameters that should be used to register the calculation.
        """

        # run is always used to register but is never an input parameter
        parameters_to_register = []
        # workflow_name is also used to register but this implemented within
        # the _register_calculation method

        table_columns = cls.database_table.get_column_names()

        # as an extra, we need to check for relations and add also check for
        # "_id" added on to the name in case we want to register the new entry
        # with this relation. An example of this is "diffusion_analysis_id"
        # which is a related object and column.
        for field in cls.database_table._meta.get_fields():
            if field.is_relation:  # and isinstance(ForeignKey)
                # we append BOTH but only one should be found during registration
                table_columns.append(field.name)
                table_columns.append(f"{field.name}_id")

        for parameter in cls.parameter_names:
            if parameter in table_columns:
                parameters_to_register.append(parameter)

        # put in alphabetical order for consistent results
        parameters_to_register.sort()

        return parameters_to_register

    @classmethod
    def _register_calculation(cls, **kwargs) -> Calculation:
        """
        If the workflow is linked to a calculation table in the Simmate database,
        this adds the flow run to the database.

        Parameters passed should be deserialized and cleaned.

        This method should not be called directly as it is used within the
        `run_prefect_cloud` method and `load_input_and_register` task.
        """

        # grab the registration kwargs from the parameters provided and then
        # convert them to a python object format for the database method
        register_kwargs = {
            key: kwargs.get(key, None) for key in cls._parameters_to_register
        }
        register_kwargs_cleaned = cls._deserialize_parameters(
            add_defaults=False, **register_kwargs
        )

        # SPECIAL CASE: The exception to the above is with SOURCE, which needs
        # to be in a JSON-serialized form for the database
        if "source" in register_kwargs_cleaned:
            register_kwargs_cleaned["source"] = cls._serialize_parameters(
                source=kwargs.get("source", None)
            )["source"]
            # !!! This is a hacky bug fix that needs refactored

        # SPECIAL CASE: for customized workflows we need to convert the inputs
        # back to json before saving to the database.
        if "workflow_base" in register_kwargs_cleaned:
            parameters_serialized = cls._serialize_parameters(**register_kwargs_cleaned)
            calculation = cls.database_table.from_run_context(
                workflow_name=cls.name_full,
                workflow_version=cls.version,
                **parameters_serialized,
            )
        else:
            # load/create the calculation for this workflow run
            calculation = cls.database_table.from_run_context(
                workflow_name=cls.name_full,
                workflow_version=cls.version,
                **register_kwargs_cleaned,
            )

        return calculation

    # -------------------------------------------------------------------------
    # Methods that hanlde serialization and deserialization of input parameters.
    # -------------------------------------------------------------------------

    @classmethod
    def _update_with_defaults(cls, parameter_dict: dict) -> dict:
        """
        updates a dictoinary default input parameters if they were provided
        """
        for parameter_name, default_value in cls.parameter_defaults.items():
            if (
                parameter_dict.get(parameter_name, None) is None
                and default_value is not None
            ):
                parameter_dict[parameter_name] = default_value

    @classmethod
    def _serialize_parameters(cls, **parameters) -> dict:
        """
        Converts input parameters to json-sealiziable objects that Prefect can
        use.

        This method should not be called directly as it is used within the
        run_prefect_cloud() method.
        """

        # TODO: consider moving this into prefect's core code as a contribution.
        # This alternatively might be a pydantic contribution

        # Because many flows allow object-type inputs (such as structure object),
        # we need to serialize these inputs before scheduling them with prefect
        # cloud. To do this, I only check for two potential methods:
        #     as_dict
        #     to_dict
        # These were chosen based on common pymatgen methods, but I may change
        # this in the future. As another alternative, I could also cloudpickle
        # input objects when they are not JSON serializable.
        # OPTIMIZE: Prefect current tries to JSON serialize within their
        # client.create_flow_run method. In the future, we may want to move
        # this functionality there.

        parameters_serialized = {}
        for parameter_key, parameter_value in parameters.items():

            try:
                json.dumps(parameter_value)
            except TypeError:

                # Special cases
                if parameter_key == "directory":
                    # convert Path to str
                    parameter_value = str(parameter_value)
                elif parameter_key == "source" and isinstance(parameter_value, dict):
                    # recursive call to this function
                    parameter_value = cls._serialize_parameters(**parameter_value)
                elif parameter_key == "composition":
                    # convert Composition to str
                    parameter_value = str(parameter_value)

                # preferred serializiation
                elif hasattr(parameter_value, "as_dict"):
                    parameter_value = parameter_value.as_dict()
                elif hasattr(parameter_value, "to_dict"):
                    parameter_value = parameter_value.to_dict()

                # workflow_base and input_parameters are special cases that
                # may require a refactor (for customized workflows)
                elif parameter_key == "workflow_base":
                    parameter_value = parameter_value.name_full
                elif parameter_key == "input_parameters":
                    # recursive call to this function
                    parameter_value = cls._serialize_parameters(**parameter_value)

                else:
                    parameter_value = cloudpickle.dumps(parameter_value)
            parameters_serialized[parameter_key] = parameter_value
        return parameters_serialized

    @classmethod
    def _deserialize_parameters(
        cls,
        add_defaults: bool = True,
        **parameters,
    ) -> dict:
        """
        converts all parameters to appropriate python objects
        """

        from simmate.toolkit import Composition, Structure
        from simmate.toolkit.diffusion import MigrationHop, MigrationImages

        parameters_cleaned = parameters.copy()

        #######
        # SPECIAL CASE: customized workflows have their parameters stored under
        # "input_parameters" instead of the base dict
        # THIS INVOLVES A RECURSIVE CALL TO THIS SAME METHOD
        if "workflow_base" in parameters.keys():
            # This is a non-modular import that can cause issues and slower
            # run times. We therefore import lazily.
            from simmate.workflows.utilities import get_workflow

            # Make sure we have a workflow object
            parameters_cleaned["workflow_base"] = (
                get_workflow(parameters["workflow_base"])
                if isinstance(parameters["workflow_base"], str)
                else parameters["workflow_base"]
            )
            # Make a recursive call for the input parameters
            parameters_cleaned["input_parameters"] = Workflow._deserialize_parameters(
                **parameters["input_parameters"]
            )
            return parameters_cleaned
        #######

        # We populate this dictionary using the default input parameters if
        # they were provided.
        if add_defaults:
            cls._update_with_defaults(parameters_cleaned)

        # The remaining checks look to intialize input to toolkit objects using
        # the from_dynamic methods.
        # OPTIMIZE: if I have proper typing and parameter itrospection, I could
        # potentially grab the from_dynamic method on the fly -- rather than
        # doing these repeated steps here.
        parameter_mappings = {
            "structure": Structure,
            "composition": Composition,
            "migration_hop": MigrationHop,
            "migration_images": MigrationImages,
            "supercell_start": Structure,
            "supercell_end": Structure,
        }

        for parameter, target_class in parameter_mappings.items():
            if parameter in parameters.keys():
                parameter_orig = parameters.get(parameter, None)
                parameters_cleaned[parameter] = target_class.from_dynamic(
                    parameter_orig
                )

        # directory and source are two extra parameters that cant be used in the
        # mapping above because they don't have a `from_dynamic` method.
        # Note these also pull from 'parameters_cleaned' as they might have been
        # populated during registration.

        if parameters.get("directory", None):
            parameters_cleaned["directory"] = Path(parameters_cleaned["directory"])

        if parameters.get("source", None):
            # !!! are there other types I should account for? Maybe I should just
            # make this a recursive call to catch everything?
            if parameters_cleaned["source"].get("directory", None):
                parameters_cleaned["source"]["directory"] = Path(
                    parameters_cleaned["directory"]
                )

        return parameters_cleaned



# -*- coding: utf-8 -*-

import warnings

# BUG: Importing matminer featurizers print a tqdm error so we silence them
# by importing them all together here.
with warnings.catch_warnings(record=True):
    from matminer.featurizers.site import CrystalNNFingerprint
    from matminer.featurizers.structure import (
        PartialRadialDistributionFunction,
        RadialDistributionFunction,
        SiteStatsFingerprint,
    )



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .neb_all_paths_mit import Diffusion__Vasp__NebAllPathsMit
from .neb_from_endpoints_mit import Diffusion__Vasp__NebFromEndpointsMit
from .neb_from_images_mit import Diffusion__Vasp__NebFromImagesMit
from .neb_from_images_mvl_ci import Diffusion__Vasp__NebFromImagesMvlCi
from .neb_single_path_mit import Diffusion__Vasp__NebSinglePathMit



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .utilities import write_and_submit_structures

from .new_individual import StructurePrediction__Toolkit__NewIndividual
from .fixed_composition import StructurePrediction__Toolkit__FixedComposition
from .variable_nsites_composition import (
    StructurePrediction__Toolkit__VariableNsitesComposition,
)
from .binary_system import StructurePrediction__Toolkit__BinarySystem



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.electronic_structure.base_density_of_states import (
    VaspDensityOfStates,
)
from simmate.calculators.vasp.workflows.relaxation.matproj import (
    Relaxation__Vasp__Matproj,
)


class ElectronicStructure__Vasp__MatprojDensityOfStates(
    VaspDensityOfStates, Relaxation__Vasp__Matproj
):
    """
    This task is a reimplementation of pymatgen's
    [MPNonSCFSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPNonSCFSet)
    with mode="uniform".

    Calculates the band structure using Materials Project HSE settings.
    """

    incar = Relaxation__Vasp__Matproj.incar.copy()
    incar.update(
        IBRION=-1,
        LCHARG=False,
        LORBIT=11,
        LWAVE=False,
        NSW=0,
        ISYM=2,
        ICHARG=11,
        ISMEAR=-5,
        NEDOS=2001,
    )
    incar.pop("MAGMOM__smart_magmom")



# -*- coding: utf-8 -*-

"""
This defines the base "simmate" command that all other commands stem from.
"""

import logging
from pathlib import Path

import typer

from simmate.command_line.database import database_app
from simmate.command_line.utilities import utilities_app
from simmate.command_line.workflow_engine import workflow_engine_app
from simmate.command_line.workflows import workflows_app

simmate_app = typer.Typer(rich_markup_mode="markdown")


@simmate_app.callback(no_args_is_help=True)
def base_command():
    """
    This is the base command that all other Simmate commands stem from
    :fire::fire::rocket:

    ---------

    If you are a beginner to the command line, be sure to start with
    [our tutorials](https://jacksund.github.io/simmate/getting_started/overview/).
    Below you will see a list of sub-commands to try. For example, you can run `simmate
    database --help` to learn more about it.

    \n
    TIP: Many Simmate commands are long and verbose. You can use `--install-completion`
    to add ipython-like autocomplete to your shell.

    """
    # When we call the command "simmate" this is where we start, and it then
    # looks for all other functions that have the decorator "@simmate.command()"
    # to decide what to do from there.
    pass


@simmate_app.command()
def run_server():
    """
    Runs a local test server for the Simmate website interface

    While this command is running, you can then view the working website
    at http://127.0.0.1:8000/

    This command is exactly the same as running:

    `django-admin runserver --settings=simmate.configuration.django.settings`
    """

    # BUG: windows dev version is throwing issues with this code section here,
    # so I just switch to using subprocess below.
    # ---> Error while finding module specification for '__main__'
    # (ValueError: __main__.__spec__ is None)
    #
    # logging prints ugly duplicates as it sets up simmate and also a
    # static server for it.
    # logging.warning(
    #     "Seeing duplicate messages is normal and expected. "
    #     "This is because the test server sets up + tears down Simmate on a cycle."
    # )
    # from django.core.management import call_command
    # from simmate.database import connect
    # call_command("runserver")

    import subprocess

    logging.info("Setting up local test server...")
    subprocess.run(
        "django-admin runserver --settings=simmate.configuration.django.settings",
        shell=True,
    )


@simmate_app.command()
def start_project():
    """
    Creates a new folder and fills it with an example project to
    get you started with custom Simmate workflows/datatables
    """

    import logging
    import shutil

    from simmate import configuration  # needed for the filepath

    # This directory is where our template folder is located. We find this
    # by looking at the import path to see where python installed it.
    config_directory = Path(configuration.__file__).absolute().parent

    # We add on "project_template" to this file path as that's our full template
    template_directory = config_directory / "example_project"

    # grab the full path to the new project for the user to see
    new_project_directory = Path.cwd() / "my_simmate_project"

    # copy the directory over
    shutil.copytree(template_directory, new_project_directory)

    # Let the user know what we did and how to continue.
    logging.info(
        "Successfully made a new project! \n\n"
        f"You'll find it at {new_project_directory}\n\n"
        "Note, this project's app has not yet been named, installed, or "
        "registered yet. Follow along with our tutorial to complete these steps."
    )


# # All commands are organized into other files, so we need to manually register
# # them to our base "simmate" command here.
simmate_app.add_typer(database_app, name="database")
simmate_app.add_typer(workflows_app, name="workflows")
simmate_app.add_typer(workflow_engine_app, name="workflow-engine")
simmate_app.add_typer(utilities_app, name="utilities")



# -*- coding: utf-8 -*-

# Example job(s) [with and without subprocess.Popen]
# Example handler(s)

# Test...
# setup a working directory when given None, tempdir, and string of dir
# non-zero exit code raises error
# output compression (assert location and content)
# log output (assert naming, content, and location)
# return corrections (on and off)
# a simple SupervisedJobTask run (without handlers, with, and with monitors)
# run with and without handlers
# run with and without monitors
# terminate with a monitor
# terminate with a special-case monitor
# catch error with a non-monitor
# test max_errors limit

import shutil

import pytest

from simmate.workflow_engine import ErrorHandler, S3Workflow
from simmate.workflow_engine.s3_workflow import (
    CommandNotFoundError,
    MaxCorrectionsError,
    NonZeroExitError,
)

# ----------------------------------------------------------------------------

# make some a simple ErrorHandlers for us to test with


class AlwaysPassesHandler(ErrorHandler):
    def check(self, directory):
        return False

    def correct(self, directory):
        # this should never be entered since check() never returns an error
        raise Exception


class AlwaysFailsHandler(ErrorHandler):
    def check(self, directory):
        return True

    def correct(self, directory):
        return "ExampleCorrection"


class AlwaysPassesMonitor(AlwaysPassesHandler):
    is_monitor = True


class AlwaysFailsMonitor(AlwaysFailsHandler):
    is_monitor = True


class AlwaysPassesSpecialMonitor(AlwaysPassesMonitor):
    has_custom_termination = True

    def terminate_job(self, directory, **kwargs):
        return True


class AlwaysFailsSpecialMonitorRetry(AlwaysFailsMonitor):
    has_custom_termination = True

    def terminate_job(self, directory, **kwargs):
        return True


class AlwaysFailsSpecialMonitorNoRetry(AlwaysFailsMonitor):
    has_custom_termination = True

    def terminate_job(self, directory, **kwargs):
        return False


# ----------------------------------------------------------------------------


# @pytest.mark.prefect_db
def test_s3workflow_methods(tmp_path):
    class Customized__Testing__DummyWorkflow(S3Workflow):
        command = "echo dummy"
        use_database = False

    # for shorthand reference below
    workflow = Customized__Testing__DummyWorkflow

    assert isinstance(workflow.get_config(), dict)

    workflow.show_config()  # a print statment w. nothing else to check

    # Test basic run
    state = workflow.run(directory=tmp_path)
    result = state.result()
    assert state.is_completed()
    assert result == {"corrections": []}


def test_s3workflow_1(tmp_path):
    # run a basic task w.o. any handlers or monitoring

    class Customized__Testing__DummyWorkflow(S3Workflow):
        command = "echo dummy"
        use_database = False
        monitor = False

    output = Customized__Testing__DummyWorkflow.run_config(directory=tmp_path)

    assert output == {"corrections": []}


def test_s3workflow_2(tmp_path):
    # Make a task with error handlers, monitoring, and specific directory

    class Customized__Testing__DummyWorkflow(S3Workflow):
        command = "echo dummy"
        use_database = False
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [
            AlwaysPassesHandler(),
            AlwaysPassesMonitor(),
            AlwaysPassesSpecialMonitor(),
        ]

    # use the temporary directory
    output = Customized__Testing__DummyWorkflow.run_config(directory=tmp_path)
    assert output == {"corrections": []}


def test_s3workflow_3(tmp_path):
    # test nonzero returncodes

    class Customized__Testing__DummyWorkflow(S3Workflow):
        use_database = False
        command = "NonexistantCommand 404"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysPassesHandler()]

    pytest.raises(
        CommandNotFoundError,
        Customized__Testing__DummyWorkflow.run_config,
        directory=tmp_path,
    )

    class Customized__Testing__DummyWorkflow(S3Workflow):
        use_database = False
        command = "cd xyz123"  # will fail with "directory does not exist"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysPassesHandler()]

    pytest.raises(
        NonZeroExitError,
        Customized__Testing__DummyWorkflow.run_config,
        directory=tmp_path,
    )


def test_s3workflow_4(tmp_path):
    # testing handler-triggered failures

    class Customized__Testing__DummyWorkflow(S3Workflow):
        use_database = False
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysFailsHandler()]

    pytest.raises(
        MaxCorrectionsError,
        Customized__Testing__DummyWorkflow.run_config,
        directory=tmp_path,
    )


def test_s3workflow_5(tmp_path):
    # monitor failure

    class Customized__Testing__DummyWorkflow(S3Workflow):
        use_database = False
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysFailsMonitor()]

    pytest.raises(
        MaxCorrectionsError,
        Customized__Testing__DummyWorkflow.run_config,
        directory=tmp_path,
    )


def test_s3workflow_6(tmp_path):
    # special-monitor failure (non-terminating monitor)

    class Customized__Testing__DummyWorkflow(S3Workflow):
        use_database = False
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysFailsSpecialMonitorRetry()]

    pytest.raises(
        MaxCorrectionsError,
        Customized__Testing__DummyWorkflow.run_config,
        directory=tmp_path,
    )


def test_s3workflow_7(tmp_path):
    # check that monitor exits cleanly when retries are not allowed and no
    # workup method raises an error

    class Customized__Testing__DummyWorkflow(S3Workflow):
        use_database = False
        command = "echo dummy"
        polling_timestep = 0
        monitor_freq = 2
        error_handlers = [AlwaysFailsSpecialMonitorNoRetry()]

    result = Customized__Testing__DummyWorkflow.run_config(directory=tmp_path)
    assert len(result["corrections"]) == 1


def test_s3workflow_8(tmp_path):
    # make sure an error is raised when a file is missing

    class Customized__Testing__DummyWorkflow(S3Workflow):
        use_database = False
        command = "echo dummy"
        required_files = ["FILE_THAT_DOESNT_EXIST"]

    pytest.raises(
        Exception,
        Customized__Testing__DummyWorkflow.run_config,
        directory=tmp_path,
    )


# !!! Unitests to use with Prefect Executor
# Test as a subflow
# from prefect import flow
# @flow
# def test():
#     state = workflow.run()
#     return state.result()
# state = test(return_state=True)
# result = state.result()
# assert state.is_completed()
# assert result["directory"].exists()
# shutil.rmtree(result["directory"])



# -*- coding: utf-8 -*-

from pathlib import Path

from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.electronic_structure.plotter import DosPlotter
from pymatgen.io.vasp.outputs import Vasprun

from simmate.database.base_data_types import (
    Calculation,
    DatabaseTable,
    Structure,
    table_column,
)
from simmate.visualization.plotting import MatplotlibFigure


class DensityofStates(DatabaseTable):
    """
    The electronic density of states holds information about the range of energy
    levels that are available within a material.
    """

    class Meta:
        abstract = True

    exclude_from_summary = ["density_of_states_data"]

    archive_fields = ["density_of_states_data"]

    api_filters = dict(
        band_gap=["range"],
        energy_fermi=["range"],
        conduction_band_minimum=["range"],
        valence_band_maximum=["range"],
    )

    # !!! Consider breaking down data into...
    #   total
    #   elemental -- dict of ["total", "s", "p", "d", "f"], ["1", "-1"]
    #   orbital
    density_of_states_data = table_column.JSONField(blank=True, null=True)
    """
    A JSON dictionary holding all information for the band structure. This JSON
    is generated using pymatgen's `vasprun.complete_dos.as_dict()` and is 
    therefore currently unoptimized for small storage.
    """

    band_gap = table_column.FloatField(blank=True, null=True)
    """
    The band gap energy in eV.
    """

    energy_fermi = table_column.FloatField(blank=True, null=True)
    """
    The Fermi energy in eV.
    """

    conduction_band_minimum = table_column.FloatField(blank=True, null=True)
    """
    The conduction band minimum in eV.
    """

    valence_band_maximum = table_column.FloatField(blank=True, null=True)
    """
    The valence band maximum in eV.
    """

    # TODO: consider adding...
    # spin_polarization (float "Spin polarization at the fermi level.")
    # magnetic_ordering (Magnetic ordering of the calculation.)

    def write_output_summary(self, directory: Path):
        """
        In addition to writing the normal VASP output summary, this also plots
        the DOS to "density_of_states.png"
        """

        super().write_output_summary(directory)
        self.write_dos_diagram_plot(directory=directory)

    @classmethod
    def _from_toolkit(
        cls,
        density_of_states: CompleteDos = None,
        as_dict=False,
    ):
        # Given energy, this function builds the rest of the required fields
        # for this class as an object (or as a dictionary).
        data = (
            dict(
                density_of_states_data=density_of_states.as_dict(),
                band_gap=float(density_of_states.get_gap()),
                energy_fermi=density_of_states.efermi,
                conduction_band_minimum=float(density_of_states.get_cbm_vbm()[0]),
                valence_band_maximum=float(density_of_states.get_cbm_vbm()[1]),
            )
            if density_of_states
            else {}
        )

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return data if as_dict else cls(**data)

    def to_toolkit_density_of_states(self) -> CompleteDos:
        """
        Converts this DatabaseTable object into a toolkit CompleteDos, which
        has many more methods for plotting and analysis.
        """
        return CompleteDos.from_dict(self.density_of_states_data)


class DensityofStatesCalc(Structure, DensityofStates, Calculation):
    """
    Holds Structure, DensityofStates, and Calculation information. Density of state
    workflows are common in materials science, so this table defines the most
    common data that results from such workflow calculations.
    """

    class Meta:
        app_label = "workflows"

    @classmethod
    def from_vasp_run(cls, vasprun: Vasprun, as_dict: bool = False):

        density_of_states_db = cls.from_toolkit(
            structure=vasprun.structures[0],
            density_of_states=vasprun.complete_dos,
            as_dict=as_dict,
        )
        if not as_dict:
            density_of_states_db.save()
        return density_of_states_db


class DosDiagram(MatplotlibFigure):
    def get_plot(result: DensityofStates):

        # NOTE: This method should be moved to a toolkit object

        # DEV NOTE: Pymatgen only implements matplotlib for their DOS
        # at the moment, but there are two scripts location elsewhere that can
        # outline how this can be done with Plotly:
        # https://plotly.com/python/v3/ipython-notebooks/density-of-states/
        # https://github.com/materialsproject/crystaltoolkit/blob/main/crystal_toolkit/components/bandstructure.py

        plotter = DosPlotter()
        complete_dos = result.to_toolkit_density_of_states()

        # Add the total density of States
        plotter.add_dos("Total DOS", complete_dos)

        # add element-projected density of states
        plotter.add_dos_dict(complete_dos.get_element_dos())

        # If I want plots for individual orbitals
        # for site in vasprun.final_structure:
        #     spd_dos = vasprun.complete_dos.get_site_spd_dos(site)
        #     plotter.add_dos_dict(spd_dos)

        # NOTE: get_dos_dict may be useful in the future

        plot = plotter.get_plot()
        return plot


# register all plotting methods to the database table
for _plot in [DosDiagram]:
    _plot.register_to_class(DensityofStates)



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import Calculation, table_column


class BinarySystemSearch(Calculation):
    class Meta:
        app_label = "workflows"

    chemical_system = table_column.CharField(max_length=10, null=True, blank=True)
    subworkflow_name = table_column.CharField(max_length=200, null=True, blank=True)
    subworkflow_kwargs = table_column.JSONField(default=dict, null=True, blank=True)
    max_atoms = table_column.IntegerField(null=True, blank=True)
    max_stoich_factor = table_column.IntegerField(null=True, blank=True)
    singleshot_sources = table_column.JSONField(default=list, null=True, blank=True)



# -*- coding: utf-8 -*-

import logging

import cloudpickle  # needed to serialize Prefect workflow runs and tasks
from rich import print

from simmate.workflow_engine.execution.database import WorkItem


class SimmateExecutor:
    """
    Sets up a connection to the queue database. Unlike normal executors,
    this does not set up any workers -- you must launch Worker instances
    elsewhere. It's primary role is to connect to the queue database
    and generate futures for workers. Therefore, think of the Executor
    as how you SUBMIT tasks and then the Worker is how you RUN jobs. You
    need both classes to have the setup working properly.

    Only use this Executor when Dask can't solve your problem! It's main
    use it to bypass university HPC cluster's firewalls because here worker
    signals are one-directional -- that is they query a database and there
    is never a signal sent to the worker like other executors do. Thus
    we can have workers anywhere we'd like as long as they have access
    to internet - so even multiple HPC clusters will work. At the moment,
    the executor has no idea how many workers exist and their state. I may
    add a "worker heartbeat" table to the queue database for the executor
    to read and run managerial tasks based off though.
    """

    # This class is modeled after the following...
    # https://github.com/python/cpython/blob/master/Lib/concurrent/futures/thread.py
    # https://docs.python.org/3/library/concurrent.futures.html
    # from concurrent.futures import Executor # No need to inherit at the moment

    @staticmethod
    def submit(
        fxn: callable,
        *args,
        tags: list[str] = [],
        **kwargs,
    ) -> WorkItem:

        # The *args and **kwargs input separates args into a tuple and kwargs into
        # a dictionary for me, which makes their storage very easy!

        # make the WorkItem where all of the provided inputs are pickled and
        # save the workitem to the database.
        # Pickling is just converting them to a byte string format
        # No lock is needed to do this because adding a new row is handled
        # by the database with ease, even if some different Executor is
        # adding another WorkItem at the same time.
        # TODO - should I put pickling in a "try" in case it fails?
        workitem = WorkItem.objects.create(
            fxn=cloudpickle.dumps(fxn),
            args=cloudpickle.dumps(args),
            kwargs=cloudpickle.dumps(kwargs),
            tags=tags,  # should be json serializable already
        )

        # and return the workitem/future for use
        return workitem

    @staticmethod
    def wait(workitems: list[WorkItem]):
        """
        Waits for all futures to complete before returning a list of their results
        """
        # If a dictionary of {key1: future1, key2: future2, ...} is given,
        # then we return a dictionary of which futures replaced by results.
        # NOTE: this is really for compatibility with Prefect's FlowRunner.
        if isinstance(workitems, dict):
            logging.info("waiting for workflows to finish")
            import time

            time.sleep(10)
            for key, workitem in workitems.items():
                print((key, workitem, workitem.pk, workitem.done()))
                # print(future.result())
            return {key: workitem.result() for key, workitem in workitems.items()}
        # otherwise this is a list of futures, so return a list of results
        else:
            return [workitem.result() for workitem in workitems]

    # -------------------------------------------------------------------------
    # These methods are for managing and monitoring the queue
    # I attach these directly to the Executor rather than having a separate
    # DjangoQueue class that inherits from python's Queue module.
    # If there is a good reason to make a separate class in the future,
    # I can start from these methods here and the following link:
    # https://docs.python.org/3/library/queue.html
    # -------------------------------------------------------------------------

    @staticmethod
    def queue_size() -> int:
        """
        Return the approximate size of the queue.
        """
        # Count the number of WorkItem(s) that have a status of PENDING
        # !!! Should I include RUNNING in the count? If so I do that with...
        #   from django.db.models import Q
        #   ...filter(Q(status="P") | Q(status="R"))
        queue_size = WorkItem.objects.filter(status="P").count()
        return queue_size

    @staticmethod
    def clear_queue(are_you_sure: bool = False):
        """
        Empties the WorkItem database table and delete everything. This will
        not stop the workers if they are in the middle of a job though.
        """
        # Make sure the user ment to do this, otherwise raise an exception
        if not are_you_sure:
            raise Exception(
                "Are you sure you want to do this? it deletes all of your queue"
                "data and you can't get it back. If so, set are_you_sure=True."
            )
        else:
            WorkItem.objects.all().delete()

    @staticmethod
    def clear_finished(are_you_sure: bool = False):
        """
        Empties the WorkItem database table and delete everything. This will
        not stop the workers if they are in the middle of a job though.
        """
        # Make sure the user ment to do this, otherwise raise an exception
        if not are_you_sure:
            raise Exception
        else:
            WorkItem.objects.filter(status="F").delete()

    @staticmethod
    def show_error_summary():

        errored_jobs = WorkItem.objects.filter(status="E").all()

        if not errored_jobs:
            print("No errored jobs found. Everything looks good.")
            return

        print("ID | ERROR")

        for job in errored_jobs:
            try:
                job.result()
            except Exception as error:
                print(f"{job.id} | {error}")

    # -------------------------------------------------------------------------
    # Extra methods to add if I want to be consistent with other Executor classes
    # -------------------------------------------------------------------------

    # @staticmethod
    # def map(fxn, iterables, timeout=None, chunksize=100):  # TODO
    #     # chunksize indicates how many to add at one
    #     # iterables is a list of (*args, **kwargs)
    #     # add many fn(*args, **kwargs) to queue

    #     # TODO -- This is not supported at the moment. I should use the
    #     # .bulk_create method to do this in the future:
    #     # https://docs.djangoproject.com/en/3.1/ref/models/querysets/#bulk-create

    #     # raise an error to ensure user sees this isn't supported yet.
    #     raise Exception("This method is not supported yet")

    # @staticmethod
    # def shutdown(wait=True, cancel_futures=False):  # TODO
    #     # whether to wait until the queue is empty
    #     # whether to cancel futures and clear database
    #     pass



# -*- coding: utf-8 -*-

import logging

import numpy

from simmate.toolkit import Composition
from simmate.toolkit.validators.fingerprint.base import FingerprintValidator

from .featurizers import CrystalNNFingerprint as cnnf
from .featurizers import SiteStatsFingerprint as pssf  # PartialsSiteStatsFingerprint


class PartialCrystalNNFingerprint(FingerprintValidator):
    @staticmethod
    def get_featurizer(
        composition: Composition,
        stat_options: list[str] = ["mean", "std_dev", "minimum", "maximum"],
        **crystalnn_options,
    ):

        # BUG: waiting on this pSS to be added to matminer. See the following PR:
        #   https://github.com/hackingmaterials/matminer/pull/809
        logging.warning(
            "PartialsSiteStatsFingerprint current acts as a SiteStatsFingerprint. "
            "We are waiting on a new MatMiner release to fix this issue."
        )

        # make the matminer featurizer object using the provided settings
        if crystalnn_options:
            sitefingerprint_method = cnnf(**crystalnn_options)
        # if no settings were given, we will use those from the Materials Project
        #   https://docs.materialsproject.org/user-guide/structure-similarity/
        else:
            # We change x_diff_weight to 3 here because we want the coordation
            # to be atom-dependent (best we can do is oxidation-state dependent here)
            sitefingerprint_method = cnnf.from_preset(
                "ops", distance_cutoffs=None, x_diff_weight=3
            )

        # now that we made the sitefingerprint_method, we can input it into the
        # structurefingerprint_method which finishes up the featurizer
        featurizer = pssf(sitefingerprint_method, stats=stat_options)

        # for this specific featurizer, you're supposed to use the .fit() function
        # with an example structure but really, all that does is get a list of
        # unique elements - so we can make that with a composition object
        featurizer.elements_ = numpy.array(
            [element.symbol for element in composition.elements]
        )

        return featurizer



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .steadystate_source import SteadystateSource
from .fixed_composition import FixedCompositionSearch
from .variable_nsites_composition import VariableNsitesCompositionSearch
from .binary_system import BinarySystemSearch



# -*- coding: utf-8 -*-

from .badelf import PopulationAnalysis__Bader__Badelf
from .bader import PopulationAnalysis__Bader__Bader
from .combine_chgcars import PopulationAnalysis__Bader__CombineChgcars



# -*- coding: utf-8 -*-

import shutil
from pathlib import Path

from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from simmate.calculators.vasp.inputs import Incar, Kpoints, Poscar, Potcar
from simmate.toolkit import Structure
from simmate.workflow_engine import S3Workflow


def get_default_parallel_settings():
    """
    We load the user's default parallel settings from
        ~/simmate/vasp/INCAR_parallel_settings
    If this file doesn't exist, then we just use an empty dictionary.
    """
    settings_filename = Path.home() / "simmate" / "vasp" / "INCAR_parallel_settings"
    if settings_filename.exists():
        return Incar.from_file(settings_filename)
    else:
        return {}


class VaspWorkflow(S3Workflow):

    _parameter_methods = S3Workflow._parameter_methods + ["_get_clean_structure"]

    required_files = ["INCAR", "POTCAR", "POSCAR"]

    command: str = "vasp_std > vasp.out"
    """
    The command to call vasp, which is typically vasp_std. To ensure error
    handlers work properly, make sure your command has "> vasp.out" at the end.
    """
    # TODO: add support for grabbing a user-set default from their configuration
    # TODO: add auto check for vasp.out ending

    incar: dict = None
    """
    This sets the default vasp settings from a dictionary. This is the one thing
    you *must* set when subclassing VaspWorkflow. An example is:
        
    ``` python
      incar = dict(NSW=0, PREC="Accurate", KSPACING=0.5)
    ```
    """

    incar_parallel_settings: dict = get_default_parallel_settings()
    """
    The parallel settings to add on to the base incar. These should not effect 
    the calculation result in any way (only how fast it completes), but they 
    are still selected based on the computer specs and what runs fastest on it.
    Therefore, these settings are loaded from ~/simmate/vasp/INCAR_parallel_settings
    by default and adding to this file should be the preferred method for updating
    these settings.
    """

    # TODO: add options for poscar formation
    # add_selective_dynamics=False
    # add_velocities=False
    # significant_figures=6 --> rounding issues? what's the best way to do this?

    kpoints = None
    """
    (experimental feature)
    The KptGrid or KptPath generator used to create the KPOINTS file. Note,
    this attribute is optional becuase VASP supports setting Kpts by adding
    KSPACING to the INCAR. If KSPACING is set in the INCAR, we ignore whatever 
    is set here.
    """
    # TODO - KptGrid is just a float for now, so there's no typing here.

    functional: str = None
    """
    This directs which Potcar files to grab. You would set this to a string
    of what you want, such as "PBE", "PBE_GW", or "LDA".
    """

    potcar_mappings: dict = None
    """
    This is an optional parameter to override Simmate's default selection of
    potentials based off of the functional chosen. The defaults are located
    in simmate.calculators.vasp.inputs.potcar_mappings. You can supply your
    own mapping dictionary or update the specific potentials you'd like. 
    For example:
    
    ``` python
      from simmate.calculators.vasp.inputs.potcar_mappings import PBE_ELEMENT_MAPPINGS
      element_mappings = PBE_ELEMENT_MAPPINGS.copy()  # don't forget to copy!
      element_mappings.update({"C": "C_h"})  # if you wish to update any
    ```
    
    or if you only use Carbon and don't care about other elements...
    
    ``` python
      element_mappings = {"C": "C_h"}
    ```
    
    Read more on this inside the Potcar class and be careful with updating!
    """

    standardize_structure: str | bool = False
    """
    In some cases, we may want to standardize the structure during our setup().
    
    This means running symmetry analysis on the structure in order to reduce
    the symmetry and also converting it to some standardized form. There
    are three different forms to choose from and thus 3 different values that
    `standardize_structure` can be set to:
    
    - `primitive`: for the standard primitive unitcell
    - `conventional`: for the standard conventional unitcell
    - `primitive-LLL`: for the standard primitive unitcell that is then LLL-reduced
    - `False`: this is the default and will disable this feature
    
    We recommend using `primitive-LLL` when the smallest possible and most cubic
    unitcell is desired.
    
    We recommend using `primitive` when calculating band structures and 
    ensuring we have a standardized high-symmetry path. Note, existing band
    structure workflows automatically adjust this value for you.
    
    To control the tolerances used to symmetrize the structure, you can use the
    symmetry_precision and angle_tolerance attributes.
    
    By default, no standardization is applied.
    """

    symmetry_precision: float = 0.01
    """
    If standardize_structure=True, then this is the cutoff value used to determine
    if the sites are symmetrically equivalent. (in Angstroms)
    """

    angle_tolerance: float = 0.5
    """
    If standardize_structure=True, then this is the cutoff value used to determine
    if the angles between sites are symmetrically equivalent. (in Degrees)
    """

    @classmethod
    def _get_clean_structure(
        cls,
        structure: Structure,
        standardize_structure: str | bool = None,
        symmetry_precision: float = None,
        angle_tolerance: float = None,
        **kwargs,
    ) -> Structure:
        """
        Uses the class attribute for `standardize_structure`. If this is set
        to any of the listed modes, then the structure unitcell is converted
        using the proper methods.

        Note, this method is typically called within `setup` before any input
        files are written. You should never have to call it directly.
        """

        # See if these values were provided, or default to class attribute.
        # We rename this to "standardize mode" to accurately descibe the
        # variable if it is set
        standardize_mode = standardize_structure or cls.standardize_structure
        symmetry_precision = symmetry_precision or cls.symmetry_precision
        angle_tolerance = angle_tolerance or cls.angle_tolerance

        # if standardize_structure is not requested, then we just return the
        # orignal input structure.
        if not standardize_mode:

            return structure

        # Make sure the user
        if standardize_mode not in ["primitive-LLL", "conventional", "primitive"]:
            raise Exception(
                f"Standardization mode {standardize_mode} is not supported."
            )

        # Both cleaning steps start with looking at the symmetry analysis
        sym_finder = SpacegroupAnalyzer(
            structure,
            symprec=symmetry_precision,
            angle_tolerance=angle_tolerance,
        )

        # If requested, we convert to the LLL-reduced unit cell, which aims to
        # be as cubic as possible.
        if standardize_mode == "primitive-LLL":
            structure_prim = sym_finder.get_primitive_standard_structure()
            structure_cleaned = structure_prim.copy(sanitize=True)
            return structure_cleaned

        # For band structures, we need to make sure the structure is in the
        # standardized primitive form.
        if standardize_mode == "primitive":
            structure_cleaned = sym_finder.get_primitive_standard_structure(
                international_monoclinic=False,
            )
            # check for pymatgen bugs here
            check_for_standardization_bugs(structure, structure_cleaned)
            return structure_cleaned

        # lastly is the conventional standard, which is used the least.
        if standardize_mode == "conventional":
            structure_cleaned = sym_finder.get_conventional_standard_structure()
            return structure_cleaned

    @classmethod
    def setup(cls, directory: Path, structure: Structure, **kwargs):

        # run cleaning and standardizing on structure (based on class attributes)
        structure_cleaned = cls._get_clean_structure(structure, **kwargs)

        # write the poscar file
        Poscar.to_file(structure_cleaned, directory / "POSCAR")

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file
        incar = Incar(**cls.incar) + Incar(**cls.incar_parallel_settings)
        incar.to_file(
            filename=directory / "INCAR",
            structure=structure_cleaned,
        )

        # if KSPACING is not provided in the incar AND kpoints is attached to this
        # class instance, then we write the KPOINTS file
        if cls.kpoints and ("KSPACING" not in cls.incar):
            Kpoints.to_file(
                structure_cleaned,
                cls.kpoints,
                directory / "KPOINTS",
            )

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure_cleaned.composition.elements,
            cls.functional,
            directory / "POTCAR",
            cls.potcar_mappings,
        )

    @classmethod
    def setup_restart(cls, directory: Path, **kwargs):
        """
        From a working directory of a past calculation, sets up for the calculation
        to be restarted. For relaxations/dynamics this involved just copying
        the poscar to the contcar.
        """

        # establish filenames
        poscar_filename = directory / "POSCAR"
        poscar_orig_filename = directory / "POSCAR_original"
        contcar_filename = directory / "CONTCAR"
        stopcar_filename = directory / "STOPCAR"

        # TODO:
        # make an archive of the directory before we start editting files
        # make_error_archive(directory)
        # add these changes to the simmate_corrections.csv

        # delete the stopcar if it exists
        if stopcar_filename.exists():
            stopcar_filename.unlink()

        # copy poscar to a new file
        shutil.move(poscar_filename, poscar_orig_filename)

        # then CONTCAR over to the POSCAR
        shutil.move(contcar_filename, poscar_filename)

    @classmethod
    def get_config(cls):
        """
        Grabs the overall settings from the class. This is useful for printing out
        settings for users to inspect.
        """
        return {
            key: getattr(cls, key)
            for key in [
                "__module__",
                "functional",
                "incar",
                "potcar_mappings",
            ]
        }


def check_for_standardization_bugs(
    structure_original: Structure,
    structure_new: Structure,
):

    # In pymatgen, they include this code with the standardization of their
    # structures because there were several bugs in the past and they want to
    # double-check themselves. I'm still using their code to standardize
    # my structures, so I should make this check too.

    vpa_old = structure_original.volume / structure_original.num_sites
    vpa_new = structure_new.volume / structure_new.num_sites

    if abs(vpa_old - vpa_new) / vpa_old > 0.02:
        raise ValueError(
            "Standardizing failed! Volume-per-atom changed... "
            f"old: {vpa_old}, new: {vpa_new}"
        )

    sm = StructureMatcher()
    if not sm.fit(structure_original, structure_new):
        raise ValueError("Standardizing failed! Old structure doesn't match new.")



# -*- coding: utf-8 -*-

from pymatgen.analysis.elasticity.strain import Strain, convert_strain_to_deformation

from simmate.toolkit import Structure
from simmate.toolkit.creators.vector import UniformlyDistributedVectors
from simmate.toolkit.transformations.base import Transformation


class LatticeStrain(Transformation):
    """
    Applies a large amount of strain to a lattice by shifting lattice vectors.
    The volume is then scaled back to the original volume. This is meant to
    break symmetries and allow a structure to relax.
    """

    #!!! should I mutate atomic sites too? USPEX is unclear if they do this
    # in addition to lattice strain. If so, I should look at the
    # CoordinateMutation transfomation class

    # strain is applied to the lattice and then the lattice is scaled back
    # to target volume
    # https://uspex-team.org/static/file/CPC-USPEX-2006.pdf

    # It looks like structure.apply_strain() is simply scaling each lattice
    # vector and not exactly what we'd like. Instead, I found
    # pymatgen.analysis.elasticity.strain module
    # https://pymatgen.org/pymatgen.analysis.elasticity.strain.html
    # Here, my understanding is that a strain matrix is not equivalent to a
    # deformation matrix. Therefore, I will create a Strain object, then convert
    # it Deformation object, then apply it to the structure

    io_scale = "one_to_one"
    ninput = 1
    allow_parallel = False

    @staticmethod
    def apply_transformation(
        structure: Structure,
        fixed_volume: float,
        max_attempts=100,
    ):

        # after straining the lattice, we need to scale it to a fixed volume
        fixed_volume = fixed_volume

        # following along with USPEX paper, we need to first establish boundries
        # for the symmetric strain matrix
        # when generating components for this matrix, they randomly select
        # values between -1 and 1 in a guassian distribution
        component_generator = UniformlyDistributedVectors(min_value=-1, max_value=1)

        # first we need 6 strain matrix components
        components = component_generator.new_vector(size=6)

        # next we need to assemble the strain matrix using these components
        #!!! is there a better way to do this? I'm just using equation (3) in
        # the USPEX paper
        strain_matrix = [
            [1 + components[0], components[5] / 2, components[4] / 2],
            [components[5] / 2, 1 + components[1], components[3] / 2],
            [components[4] / 2, components[3] / 2, 1 + components[2]],
        ]

        # now let's switch to pymatgen functionality

        # first make the Strain object
        strain = Strain(strain_matrix)

        # now convert it to a deformation object
        #!!! do I want the shape to be 'upper' or 'symmetric'??
        deformation = convert_strain_to_deformation(strain, shape="symmetric")

        # deform the structure
        new_structure = deformation.apply_to_structure(structure)

        # scale the lattice volume back to a fixed value
        new_structure.scale_lattice(fixed_volume)

        return new_structure



# -*- coding: utf-8 -*-

import pytest
from pandas import DataFrame

from simmate.database.base_data_types import Dynamics, DynamicsIonicStep
from simmate.toolkit import Structure

# from pymatgen.io.vasp.outputs import Vasprun


@pytest.mark.django_db
def test_static_energy_table(structure):

    # test writing columns
    Dynamics.show_columns()
    DynamicsIonicStep.show_columns()

    # test writing to database
    structure_db = Dynamics.from_run_context(
        run_id="example-id-123",
        workflow_name="example.test.workflow",
        workflow_version="1.2.3",
        structure=structure,
    )
    structure_db.save()

    # try grabbing the calculation again and make sure it loaded from the
    # database rather than creating a new entry
    structure_db2 = Dynamics.from_run_context(
        run_id="example-id-123",
        workflow_name="example.test.workflow",
        workflow_version="1.2.3",
    )
    assert structure_db.id == structure_db.id

    # test converting back to toolkit and ensuring the structure is the
    # same as before
    structure_new = structure_db.to_toolkit()
    assert structure == structure_new

    # test converting search results to dataframe and to toolkit
    df = Dynamics.objects.to_dataframe()
    assert isinstance(df, DataFrame)
    structures = Dynamics.objects.to_toolkit()
    assert isinstance(structures, list)
    assert isinstance(structures[0], Structure)



# -*- coding: utf-8 -*-

import time

# import pickle
import cloudpickle  # needed to serialize Prefect workflow runs and tasks

from simmate.database.base_data_types import DatabaseTable, table_column

# BUG: I have this database table within a module that calls "database.connect"
# at a higher level... Will this cause circular import issues?


class WorkItem(DatabaseTable):
    """
    A WorkItem is a future-like

    The name WorkItem comes from the naming convention use here:
    https://github.com/python/cpython/blob/master/Lib/concurrent/futures/thread.py

    There is a repo that goes one step further and makes a PickleField for django
    but this looks over the top for what I need -- even though django-picklefield
    is only one file! Still, I can use it as a reference if I ever run into bugs
    https://github.com/gintas/django-picklefield
    """

    class Meta:
        app_label = "workflow_engine"

    tags = table_column.JSONField(default=list)
    """
    List of tags to submit the task with, which helps with submitting workers
    for a specific type of task/workflow. (e.g. ["simmate", "custom"])
    """

    # These states were originally based on the python queue module, but we
    # updated them to match Prefect states that have more flexibility:
    # https://docs.prefect.io/concepts/states/
    class StatusOptions(table_column.TextChoices):
        PENDING = "P"
        RUNNING = "R"
        CANCELLED = "C"
        ERRORED = "E"
        FINISHED = "F"

    # TODO -- I should consider indexing this column for speed because it's
    # the most queried column by far.
    status = table_column.CharField(
        max_length=1,
        choices=StatusOptions.choices,
        default=StatusOptions.PENDING,
    )
    """
    the status/state of the workitem
    """

    command_not_found_failures = table_column.IntegerField(default=0)
    """
    Keeps track of the special-case "CommandNotFound" error that is hidden by
    workers. This keeps a tally of how many Workers that this task is triggering
    to shut down. This is important because it helps prevent total cluster
    shutdown from a single typo in a command. By default, workers will label
    a task as problematic if 2 workers give CommandNotFound errors.
    """

    # For serialization, I just use the pickle module, but in the future, I may
    # want to do a priority of MsgPk >> JSON >> Pickle. I could check if the given
    # object(s) has a serialize() method, check if has a to_json() method, and then
    # as a last resort just pickle the object.
    # https://docs.python.org/3/library/pickle.html
    #
    # Pickled objects are just written as byte strings, so I stored them in django
    # as a BinaryField which accepts bytes.
    # https://docs.djangoproject.com/en/3.1/ref/models/fields/#binaryfield

    fxn = table_column.BinaryField()
    """
    The function to be called, which is serialized into a binary format.
    """

    args = table_column.BinaryField(default=cloudpickle.dumps([]))
    """
    positional arguments to be passed into fxn
    """

    kwargs = table_column.BinaryField(default=cloudpickle.dumps({}))
    """
    keyword arguments to be passed into fxn
    """

    result_binary = table_column.BinaryField(blank=True, null=True)
    """
    the output of fxn(*args, **kwargs)
    """

    source = None
    """
    Source column is not needed so setting this to None disable the column
    """

    # TODO: Consider creating a separate table for Worker and linking it to this
    # table via a foreign key.
    # worker_id = table_column.CharField(max_length=50, blank=True, null=True)

    # -------------------------------------------------------------------------
    # The methods below turn this into a future-like object
    # These methods are based on:
    #   https://docs.python.org/3/library/concurrent.futures.html
    # Note in the methods below, that I make calls to WorkItem instead of self.
    # This is to ensure I can select_for_update() and also ensure I am
    # grabbing the most recent information.
    # -------------------------------------------------------------------------

    def cancel(self) -> bool:
        """
        Attempt to cancel the call. If the call is currently being executed or
        finished running and cannot be cancelled then the method will return
        False, otherwise the call will be cancelled and the method will return
        True.
        """
        # Query the WorkItem, lock it for editting, and check the status.
        workitem = WorkItem.objects.select_for_update().get(pk=self.pk)

        # check if the status is *not* PENDING
        if workitem.status != "P":
            # if so, the job is already running or finished, in which case
            # we can't cancel it.
            return False

        else:
            # If it is still pending, we can go ahead and cancel it.
            # This does not delete the task from the queue database though
            workitem.status = "C"
            workitem.save()
            return True

    def is_cancelled(self) -> bool:
        """
        Return True if the call was successfully cancelled.
        """
        # I don't use a lock to check the status here
        workitem = WorkItem.objects.only("status").get(pk=self.pk)

        # check the status and indicate whether it is CANCELED or not
        if workitem.status == "C":
            return True
        else:
            return False

    def is_running(self) -> bool:
        """
        Return True if the call is currently being executed and cannot be cancelled.
        """
        # I don't use a lock to check the status here
        workitem = WorkItem.objects.only("status").get(pk=self.pk)

        # check the status and indicate whether it is RUNNING or not
        if workitem.status == "R":
            return True
        else:
            return False

    def is_done(self) -> bool:
        """
        Return True if the call was successfully cancelled or finished running.
        """
        # I don't use a lock to check the status here
        workitem = WorkItem.objects.only("status").get(pk=self.pk)

        # check the status and indicate whether it is FINISHED or CANCELED
        if workitem.status == "F" or workitem.status == "C":
            return True
        else:
            return False

    def result(self, timeout: float = None, sleep_step: float = 0.1) -> any:
        """
        Return the value returned by the call. If the call hasn’t yet completed
        then this method will wait up to timeout seconds. If the call hasn’t
        completed in timeout seconds, then a concurrent.futures.TimeoutError
        will be raised. timeout can be an int or float. If timeout is not
        specified or None, there is no limit to the wait time.

        If the future is cancelled before completing then CancelledError
        will be raised.

        If the call raised, this method will raise the same exception.
        """
        # if no timeout was set, use infinity so we wait forever.
        if not timeout:
            timeout = float("inf")

        # Loop endlessly until the job completes or we timeout
        time_start = time.time()

        while (time.time() - time_start) < timeout:
            # I don't use a lock to check the status here
            workitem = WorkItem.objects.only("status", "result_binary").get(pk=self.pk)
            status = workitem.status

            if status == "F" or status == "E":  # FINISHED or ERRORED
                # grab the result, unpickle it, and return it
                result = cloudpickle.loads(workitem.result_binary)
                # if the result is an Error or Exception, raise it
                if isinstance(result, Exception):
                    raise result
                # otherwise return the result as-is
                else:
                    return result

            elif status == "C":  # CANCELED
                raise CancelledError(
                    "This item was cancelled and has no result. If this is unexpected, "
                    "be sure to check your worker logs. Misconfiguration or a `command "
                    "not found` error can be the cause of your job getting cancelled."
                )

            elif status == "P" or status == "R":  # PENDING or RUNNING
                # sleep the set amount before restarting the while loop
                time.sleep(sleep_step)

        # if the loop exits and we reached this line, then we've hit the timeout
        raise TimeoutError("The time-limit to wait for this result has been exceeded")


class CancelledError(Exception):
    pass



# -*- coding: utf-8 -*-

"""

This file is for pulling AFLOW data into the Simmate database. 

AFLOW's supported REST API can be accessed via "AFLUX API". This is a separate
python package, which is maintained at https://github.com/rosenbrockc/aflow.
Note that this not from the official AFLOW team, but it is made such that keywords
are pulled dynamically from the AFLOW servers -- any updates in AFLOW's API should
be properly handled. Also structures are loaded as ASE Atom objects, which we then
convert to pymatgen.

"""

from django.db import transaction

# This looks like the easiest way to grab all of the data -- as AFLOW doesn't
# have any good documentation on doing this.
from pymatgen.analysis.prototypes import AFLOW_PROTOTYPE_LIBRARY
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from rich.progress import track

from simmate.database.third_parties import AflowPrototype


@transaction.atomic
def load_all_prototypes():

    for prototype_data in track(AFLOW_PROTOTYPE_LIBRARY):

        # first let's grab the structure
        structure = prototype_data["snl"].structure

        # To see how many unique wyckoff sites there are we also need the
        # symmetrized structure
        structure_sym = SpacegroupAnalyzer(structure).get_symmetrized_structure()

        # Organize the data into our database format
        prototype = AflowPrototype.from_toolkit(
            structure=structure,
            mineral_name=prototype_data["tags"]["mineral"],
            aflow_id=prototype_data["tags"]["aflow"],
            pearson_symbol=prototype_data["tags"]["pearson"],
            strukturbericht_symbol=prototype_data["tags"]["strukturbericht"],
            nsites_wyckoff=len(structure_sym.wyckoff_symbols),
        )

        # and save it to our database
        prototype.save()



# -*- coding: utf-8 -*-

import logging

from numpy.random import choice

from simmate.toolkit import Composition, Structure
from simmate.toolkit.creators.lattice import RSLSmartVolume
from simmate.toolkit.creators.sites.random_wyckoff import RandomWySites
from simmate.toolkit.creators.structure.base import StructureCreator
from simmate.toolkit.creators.utils import NestedFixes
from simmate.toolkit.validators.structure import SiteDistanceMatrix


class RandomSymStructure(StructureCreator):

    # BUG: because I use MultiplicityPrimitive in RandomWySites but make a
    # convential cell below, there may be a issue with the input composition
    # vs. output composition number of sites. For example, when making a Ca2N
    # structure with spacegroup 166, the conventional cell will give a
    # structure of Ca6N3. I can then convert this to the primitive cell to get
    # Ca2N, but depending on the sym settings that I use, this might not work.
    # I don't currently have structure.get_primitive_structure() in this creator
    # because it will break the Generator workflow (which assumes unchanging
    # cell between validations).

    def __init__(
        self,
        composition: Composition,
        spacegroup_include: list[int] = range(1, 231),
        spacegroup_exclude: list[int] = [],
        lattice_generation_method=RSLSmartVolume,
        lattice_gen_options: dict = {},
        site_generation_method=RandomWySites,
        site_gen_options: dict = {},
        validator_method=SiteDistanceMatrix,
        validator_options: dict = {},
        fixindicator_method=NestedFixes,
        fixindicator_options: dict = {
            "fixes": ["new_lattice", "new_sites"],
            "cutoffs": [15, 100],
        },
        remove_failed_spacegroups: bool = False,
        cleanup: bool = True,
    ):

        # sg_include = list of spacegroups that we are interested in. Default
        # is all 230 spacegroups
        # sg_exclude = list of spacegroups that we should explicitly ignore

        # method settings
        self.remove_failed_spacegroups = remove_failed_spacegroups
        self.removed_spacegroups = []
        self.cleanup = cleanup

        # save the composition for reference
        self.composition = composition

        # establish the lattice generation method and use the specified options
        # many methods include options for spacegroup_include/exclude
        # we want to override those options with what the user indicated here
        lattice_gen_options.update(
            dict(
                composition=composition,
                spacegroup_include=spacegroup_include,
                spacegroup_exclude=spacegroup_exclude,
            )
        )
        self.lattice_generator = lattice_generation_method(**lattice_gen_options)

        # establish the site generation method and use the specified options
        # many methods include options for composition and spacegroup_include/exclude
        # we want to override those options with what the user indicated here
        site_gen_options.update(
            dict(
                composition=composition,
                spacegroup_include=spacegroup_include,
                spacegroup_exclude=spacegroup_exclude,
            )
        )
        self.site_generator = site_generation_method(**site_gen_options)

        # BUG: I assume that all lattice/site methods have three inputs
        # which will cause errors when the new classes that don't accept these
        # this input types.

        # establish the validation method and use the specified options
        # many methods include options for composition
        # we want to override that option with what the user indicated here
        validator_options.update(dict(composition=composition))
        self.validator = validator_method(**validator_options)

        # establish the FixIndicator method and use the specified options
        self.fixindicator = fixindicator_method(**fixindicator_options)

        # make a list of spacegroups that we are allowed to choose from
        self.spacegroup_options = [
            sg for sg in spacegroup_include if sg not in spacegroup_exclude
        ]

        # While a given sg_inlcude and sg_exclude were given, some spacegroup
        # may not be compatible with the lattice or site method used. For example,
        # the spacegroup 230 is not possible with a composition of MgSiO3 (1:1:3).
        # Selection of appropriate wy_sites can't be done. The user still may
        # have 230 in the spacegroup_include. This function will ensure all
        # spacegroups are compatible -- if not, self.spacegroup_options will be
        # updated by remove the spacegroup. Therefore, this step is to see
        # which spacegroups are compatible with BOTH the lattice creation method
        # and the site creation method.
        self.spacegroup_options = [
            sg
            for sg in self.spacegroup_options
            if sg in self.lattice_generator.spacegroup_options
            and sg in self.site_generator.spacegroup_options
        ]
        # !!! should I have a self.spacegroups_invalid list to keep track of
        # what's been removed?
        #
        # BUG: this requires all site and lattice creators to have a
        # spacegroup_options property.

    def create_structure(self, spacegroup=None):

        # If a spacegroup is not specified, grab a random one from our options
        # no check is done to see if the spacegroup specified is compatible
        # with the vector_generator built
        if not spacegroup:
            # randomly select a spacegroup
            spacegroup = choice(self.spacegroup_options)
        logging.info(f"Creating structure with spacegroup {spacegroup}")

        # Use the lattice generator to make a starting lattice
        lattice = self.lattice_generator.new_lattice(spacegroup)

        # Use the site generator to make a starting series of sites
        sites = self.site_generator.new_sites(spacegroup)
        # if the generator returns None - then the spacegroup is incompatible
        # with the composition
        if not sites:
            # The spacegroup is incompatible with the composition and
            # site-generator-method.
            # We do not raise an error in case this method needs to be called
            # on a guess/check basis.
            return False
        # otherwise, the generator was successful so we can unpack its output
        species, coords = sites

        # Set the valid_struct to false so we can start the loop below.
        is_valid = False
        # Keep track of the number of attempts we make to create a structure
        attempts = 0

        # we want to continue this loop until we get a valid structure
        while not is_valid:

            # add an attempt to the validator
            attempts += 1

            # since we only have sites in the asym unit, we use symmetry
            # operations to get the rest of the unitcell
            structure = Structure.from_spacegroup(spacegroup, lattice, species, coords)
            # BUG: Note that I'm not always limiting sites to the asym unit...
            # if I used RandomSites instead of RandomSymSites or RandomWySites,
            # this may give problems

            # Check that the structure is valid
            is_valid = self.validator.check_structure(structure)

            # If the validator passes, we can exit the loop!
            # This will be done automatically when 'while not is_valid' is checked

            # Otherwise, we need to try to fix the structure and check it again.
            if not is_valid:

                # see the assigned 'fix' for the validator
                fix = self.fixindicator.point_to_fix(attempts)

                if not fix:
                    # if the 'fix' is False, that means we maxed out the attempts
                    # on the validator.
                    # this indicates a total failure and the function should end
                    structure = False  # set to false indicating a failure
                    break  # exits the while loop with structure = False
                    # We do not return False yet because we want to
                    # I could return False here because we want to check for
                    # "remove_failed_spacegroups" below

                # The defualt "point_to_fix" settings have strings returned
                # and not objects. There are only a limited number of options
                # for the string returned -- e.g. "new_structure indicates
                # we making a new structure from scratch, while "new_lattice"
                # says to try changing up the lattice but keep the sites.
                elif fix == "new_structure":
                    # Use the lattice generator to make a starting lattice
                    lattice = self.lattice_generator.new_lattice(spacegroup)

                    # Use the site generator to make a starting series of sites
                    sites = self.site_generator.new_sites(spacegroup)
                    # if the generator returns None - then the spacegroup is
                    # incompatible with the composition
                    if not sites:
                        logging.warning(
                            f"Failed to create a structure using spacegroup {spacegroup}."
                        )
                        return False
                    # otherwise, the generator was successful so we can unpack
                    # its output
                    species, coords = sites

                elif fix == "new_lattice":
                    # Use the lattice generator to make a starting lattice
                    lattice = self.lattice_generator.new_lattice(spacegroup)

                elif fix == "new_sites":
                    # Use the site generator to make a starting series of sites
                    sites = self.site_generator.new_sites(spacegroup)
                    # if the generator returns None - then the spacegroup is
                    # incompatible with the composition
                    if not sites:
                        logging.warning(
                            f"Failed to create a structure using spacegroup {spacegroup}."
                        )
                        return False
                    # otherwise, the generator was successful so we can unpack
                    # its output
                    species, coords = sites

                # !!! Rather than using "new_structure", "new_lattice", and "new_sites"
                # like we do above, it would be more powerful to allow specific
                # settings of different creators or transformations. An example
                # of this would be using a `transformations.sites` class to
                # relax sites to a reasonable position. This code below is
                # a template for how this feature enhancement would look. I keep
                # it strictly as notes for now.
                #
                # fix_type = str(type(fix))
                # # if fix points to a generator, we want to use its .new_* method
                # if 'creators.structure' in fix_type:
                #     structure = fix.new_structure(spacegroup)
                #     # we need to update the lattice, species, coords
                #     # otherwise it will be rewritten above
                #     lattice = structure.lattice
                #     species = structure.species
                #     coords = structure.frac_coords.tolist()
                # elif 'creators.lattice' in fix_type:
                #     lattice = fix.new_lattice(spacegroup)
                # elif 'creators.sites' in fix_type:
                #     species, coords = fix.new_sites(spacegroup)
                # # if fix points to a transform, we want to use its apply_to_* method
                # elif 'transformations.structure' in fix_type:
                #     structure = fix.apply_to_structure(structure)
                #     # we need to update the lattice, species, coords otherwise
                #     # it will be rewritten above
                #     lattice = structure.lattice
                #     species = structure.species
                #     coords = structure.frac_coords.tolist()
                # elif 'transformations.lattice' in fix_type:
                #     lattice = fix.apply_to_lattice(structure.lattice)
                # elif 'transformations.sites' in fix_type:
                #     species, coords = fix.apply_to_sites(
                #         structure.species,
                #         structure.coords,
                #     )

        # If the structure creation failed (structure = False), then see if the
        # failed spacegroup should be removed from future choices.
        if not structure:
            logging.warning(
                f"Failed to create a structure using spacegroup {spacegroup}."
            )
            if self.remove_failed_spacegroups:
                logging.info(
                    f"Removed spacegroup {spacegroup} from options. "
                    "This will not be selected moving forward."
                )
                # add the removed spacegroup to our list for reference
                self.removed_spacegroups.append(spacegroup)
                # and remove the spacegroup from our list of options
                self.creator.spacegroup_options.remove(spacegroup)
        else:
            logging.info("Creation successful.")

        # Cleanup does 'fix-up' steps before returning the final structure.
        # This includes converting from the conventional cell to primitive as
        # well as sort the sites by electronegativity
        # Also make sure that structure != False
        if self.cleanup and structure:
            # convert the conventional unitcell to the primitive
            #!!! should I use a spacegroup analyzer instead?
            #!!! Should I LLL reduce?
            structure = structure.get_primitive_structure()
            # make sure the sites are sorted by electronegativity of their elements
            # this is important for some mutations
            structure.sort()  # this is faster than get_sorted_structure

        # BUG: In some cases, the stoichometry of the structure will be
        # different from the input composition. This happens when the structure
        # happens to have higher symmetry than the spacegroup it was made with.
        # For example, I could make a structure of Mg4Si4O12, but the primitive
        # cell has Mg2Si2O6 or MgSiO3. So we got back a structure with fewer
        # atoms than what we requested! This could cause issues down the line,
        # so we may want to check for this.
        # if structure.composition.num_atoms != composition.num_atoms:
        #     print('COMP ERROR')
        #     # restart the whole function
        #     structure = new_sample(spacegroup)
        #     # BUG: might hit a recursion depth error

        return structure



# -*- coding: utf-8 -*-

import logging
import shutil
from pathlib import Path

from simmate.toolkit.structure_prediction.evolution.database.fixed_composition import (
    FixedCompositionSearch,
)
from simmate.workflow_engine import Workflow


class StructurePrediction__Toolkit__NewIndividual(Workflow):
    """
    Generates a new individual for an evolutionary search algorithm.

    Note, this workflow should not be called directly, but instead used within
    higher level workflows (such as `fixed-composition`)
    """

    use_database = False

    @staticmethod
    def run_config(
        search_id: int,
        steadystate_source_id: int,
        directory: Path,
        **kwargs,
    ):

        search_db = FixedCompositionSearch.objects.get(id=search_id)
        source_db = search_db.steadystate_sources.get(id=steadystate_source_id)

        # Check the stop condition of the search and see if this new individual
        # is even needed. This will catch when a search ends while a new
        # individual was waiting in the queue
        if search_db.check_stop_condition():
            logging.info(
                "The search ended while this individual was in the queue. "
                "Canceling new individual."
            )
            shutil.rmtree(directory)
            return

        if source_db.is_transformation:

            transformer = source_db.to_toolkit()
            parent_ids, new_structure = transformer.apply_from_database_and_selector(
                selector=search_db.selector,
                datatable=search_db.individuals_completed,
                select_kwargs=dict(
                    ranking_column=search_db.fitness_field,
                    query_limit=200,  # Smarter way to do this...?
                ),
                validators=[search_db.validator],
            )
            source = {
                "transformation": source_db.name,
                "parent_ids": parent_ids,
            }

        elif source_db.is_creator:

            creator = source_db.to_toolkit()

            new_structure = creator.create_structure_with_validation(
                validators=[search_db.validator],
            )
            source = {
                "creator": source_db.name,
            }

        # if structure creation was successful, run the workflow for it
        if new_structure:
            state = search_db.subworkflow.run(
                structure=new_structure,
                source=source,
                directory=directory,
                **search_db.subworkflow_kwargs,
            )
            result = state.result()
            # NOTE: we tell the workflow to use the same directory. There is
            # good chance the user indicates that they want to compress the
            # folder to.



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.dynamics.base import DynamicsWorkflow
from simmate.calculators.vasp.workflows.relaxation.mit import Relaxation__Vasp__Mit


class Dynamics__Vasp__Mit(DynamicsWorkflow, Relaxation__Vasp__Mit):
    """
    This task is a reimplementation of pymatgen's
    [MITMDSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MITMDSet).

    Runs a molecular dynamics simulation using MIT Project settings. The lattice
    will remain fixed during the run.

    By default, this will run from 300-1200 Kelvin over 10,000 steps of 2
    femtoseconds, but start/end temperatures as well as step size/number can
    be adjusted when initializing this class. Note, setting parameters in the
    init is atypical for Simmate tasks, but we allow it for MD run because it
    does not affect the interopability of results -- that is, results can
    be compared accross runs regardless of temp/time.

    Provide your structure as the desired supercell, as the setup of this
    calculation does not modify your input structure.
    """

    incar = Relaxation__Vasp__Mit.incar.copy()
    incar.update(
        dict(
            # Unique to this task, we want to allow users to set these temperatures
            # but to keep with Simmate's strategy of showing all settings up-front,
            # we set these messages here.
            # TODO: consider making a "__user_input" incar tag that accepts a default
            TEBEG="Defaults to 300 but can be set by the user",  # start temperature
            TEEND="Defaults to 1200 but can be set by the user",  # end temperature
            POTIM="Defaults to 2 but can be set by the user",  # time step (in fs)
            NSW="Defaults to 10,000 but can be set by the user",  # number of steps
            #
            EDIFF__per_atom=1e-06,
            LSCALU=False,
            LCHARG=False,
            LPLANE=False,
            LWAVE=True,
            ISMEAR=0,
            NELMIN=4,
            LREAL=True,
            BMIX=1,
            MAXMIX=20,
            NELM=500,
            NSIM=4,  # same as VASP default but pymatgen sets this
            ISYM=0,  # turn off symmetry
            ISIF=0,  # only update atom sites; lattice is fixed; no lattice stress
            IBRION=0,  # turns on molecular dynamics
            KBLOCK=100,
            SMASS=0,
            ISPIN=1,  # pymatgen makes this a kwarg but we fix it to pmg's default
            PREC="Low",
            NBLOCK=1,  # same as VASP default but pymatgen sets this
        )
    )
    incar.pop("MAGMOM__smart_magmom")
    incar.pop("multiple_keywords__smart_ldau")
    incar.pop("EDIFF")
    incar.pop("ENCUT")

    # For now, I turn off all error handlers.
    # TODO
    error_handlers = []



# -*- coding: utf-8 -*-

import math

import plotly.graph_objects as plotly_go
from plotly.subplots import make_subplots

from simmate.calculators.vasp.workflows.relaxation.quality00 import (
    Relaxation__Vasp__Quality00,
)
from simmate.calculators.vasp.workflows.relaxation.quality01 import (
    Relaxation__Vasp__Quality01,
)
from simmate.calculators.vasp.workflows.relaxation.quality02 import (
    Relaxation__Vasp__Quality02,
)
from simmate.calculators.vasp.workflows.relaxation.quality03 import (
    Relaxation__Vasp__Quality03,
)
from simmate.calculators.vasp.workflows.relaxation.quality04 import (
    Relaxation__Vasp__Quality04,
)
from simmate.calculators.vasp.workflows.static_energy.quality04 import (
    StaticEnergy__Vasp__Quality04,
)
from simmate.visualization.plotting import PlotlyFigure
from simmate.workflow_engine import Workflow

# from simmate.calculators.vasp.database.relaxation import StagedRelaxation


class Relaxation__Vasp__Staged(Workflow):
    """
    Runs a series of increasing-quality relaxations and then finishes with a single
    static energy calculation.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation.vasp.quality00
        - relaxation.vasp.quality01
        - relaxation.vasp.quality02
        - relaxation.vasp.quality03
        - relaxation.vasp.quality04
        - static-energy.vasp.quality04

    This workflow is most useful for randomly-created structures or extremely
    large supercells. More precise relaxations+energy calcs should be done
    afterwards because ettings are still below MIT and Materials Project quality.
    """

    description_doc_short = "runs a series of relaxations (00-04 quality)"

    subworkflows = [
        Relaxation__Vasp__Quality00,
        Relaxation__Vasp__Quality01,
        Relaxation__Vasp__Quality02,
        Relaxation__Vasp__Quality03,
        Relaxation__Vasp__Quality04,
        StaticEnergy__Vasp__Quality04,
    ]

    @classmethod
    def run_config(
        cls,
        structure,
        command=None,
        source=None,
        directory=None,
        copy_previous_directory=False,
        **kwargs,
    ):

        # Our first relaxation is directly from our inputs.
        current_task = cls.subworkflows[0]
        state = current_task.run(
            structure=structure,
            command=command,
            directory=directory / current_task.name_full,
        )
        result = state.result()

        # The remaining tasks continue and use the past results as an input
        for i, current_task in enumerate(cls.subworkflows[1:]):
            preceding_task = cls.subworkflows[i]  # will be one before because of [:1]
            state = current_task.run(
                structure=result,  # this is the result of the last run
                command=command,
                directory=directory / current_task.name_full,
            )
            result = state.result()

        # when updating the original entry, we want to use the data from the
        # final result.
        final_result = {
            "structure": result.to_toolkit(),
            "energy": result.energy,
            "band_gap": result.band_gap,
            "is_gap_direct": result.is_gap_direct,
            "energy_fermi": result.energy_fermi,
            "conduction_band_minimum": result.conduction_band_minimum,
            "valence_band_maximum": result.valence_band_maximum,
        }
        return final_result

    @classmethod
    def get_energy_series(cls, **filter_kwargs):
        directories = (
            cls.all_results.filter(**filter_kwargs)
            .values_list("directory", flat=True)
            .all()
        )

        all_energy_series = []
        for directory in directories:
            energy_series = []
            for subflow in cls.subworkflows:
                query = subflow.database_table.objects.filter(
                    workflow_name=subflow.name_full,
                    directory__startswith=directory,
                    energy_per_atom__isnull=False,
                ).only("energy_per_atom")
                if query.exists():
                    result = query.get()
                    energy_series.append(result.energy_per_atom)
                else:
                    energy_series.append(None)
            all_energy_series.append(energy_series)

        return all_energy_series


class StagedSeriesConvergence(PlotlyFigure):
    method_type = "classmethod"

    def get_plot(
        workflow,  # Relaxation__Vasp__Staged
        **filter_kwargs,
    ):
        all_energy_series = workflow.get_energy_series(**filter_kwargs)

        figure = make_subplots(
            rows=math.ceil((len(workflow.subworkflows) - 1) / 3),
            cols=3,
        )

        for i in range(len(workflow.subworkflows) - 1):

            xs = []
            ys = []
            for energy_series in all_energy_series:

                xs.append(energy_series[i + 1])
                ys.append(energy_series[i])

            scatter = plotly_go.Scatter(
                x=xs,
                y=ys,
                mode="markers",
            )
            row = math.ceil((i + 1) / 3)
            col = (i % 3) + 1
            figure.add_trace(
                trace=scatter,
                row=row,
                col=col,
            )

            # Update xaxis properties
            figure.update_xaxes(
                title_text=f"{workflow.subworkflows[i + 1].name_full}",
                row=row,
                col=col,
            )
            figure.update_yaxes(
                title_text=workflow.subworkflows[i].name_full,
                row=row,
                col=col,
            )

        figure.update_layout(
            title="Energy per atom (eV) comparison for each stage",
            showlegend=False,
        )
        return figure


class StagedSeriesHistogram(PlotlyFigure):
    method_type = "classmethod"

    def get_plot(
        workflow,  # Relaxation__Vasp__Staged
        **filter_kwargs,
    ):

        all_energy_series = workflow.get_energy_series(**filter_kwargs)

        figure = plotly_go.Figure()

        for i in range(len(workflow.subworkflows) - 1):

            diffs = []
            for energy_series in all_energy_series:
                if energy_series[i + 1] and energy_series[i]:
                    d = energy_series[i + 1] - energy_series[i]
                    diffs.append(d)

            scatter = plotly_go.Histogram(
                x=diffs,
                name=(
                    f"{workflow.subworkflows[i].name_full} "
                    f"--> {workflow.subworkflows[i+1].name_full}"
                ),
            )
            figure.add_trace(trace=scatter)

        figure.update_layout(
            barmode="overlay",
            xaxis_title_text="Energy per atom change (eV)",
            yaxis_title_text="Structures (#)",
            bargap=0.05,
            legend=dict(
                yanchor="top",
                y=0.99,
                xanchor="left",
                x=0.01,
            ),
        )

        return figure


# register all plotting methods to the database table
for _plot in [StagedSeriesConvergence, StagedSeriesHistogram]:
    _plot.register_to_class(Relaxation__Vasp__Staged)



# -*- coding: utf-8 -*-

import logging
import subprocess
from pathlib import Path
from tempfile import mkdtemp


def start_cluster(
    nworkers: int,
    worker_command: str = "simmate workflow-engine start-worker",
):

    cluster_directory = mkdtemp(prefix="simmate-cluster-", dir=Path.cwd())

    logging.info(f"Starting up cluster in {cluster_directory}")

    # For as many workers that were requested, go through and start that many
    # subprocesses for individual workers.
    all_popens = []
    for n in range(nworkers):

        output_file = Path.cwd() / cluster_directory / f"worker_{n}.out"

        popen = subprocess.Popen(
            worker_command,
            shell=True,
            stdout=output_file.open("w"),
            stderr=output_file.open("w"),
        )
        all_popens.append(popen)

    logging.info(f"A total of {nworkers} workers have been started.")
    logging.info("Waiting until all workers workers shut down...")

    # Now just wait for the process to finish. Note we use communicate
    # instead of the .wait() method. This is the recommended method
    # when we have stderr=subprocess.PIPE, which we use above.
    [popen.communicate() for popen in all_popens]

    logging.info("All workers have terminated. Shutting down.")


# For SLURM, PBS, etc., I should refactor the strategy used by FireWorks:
#   https://github.com/materialsproject/fireworks/tree/main/fireworks/queue
#
# I tried a hacky approah using dask-jobqueue, but it was just too messy and
# buggy to use in production:
#
# from simmate.configuration.dask import get_dask_client
# from dask.distributed import as_completed
# n_workers = 5
# client = get_dask_client(n_workers=n_workers, threads_per_worker=1)
# futures = []
# for dask_id, dask_worker in client.cluster.workers.items():
#     future = client.submit(
#         subprocess.run,
#         f"simmate workflow-engine start-singleflow-worker > worker_{dask_id}.out",
#         shell=True,
#         # capture_output=True,
#         pure=False,
#         workers=[dask_id],
#     )
#     future.worker_id = dask_id  # for reference later
#     future.worker_nanny = dask_worker  # for reference later
#     futures.append(future)
# for future in as_completed(futures):
#     print("restarting")
#     future.worker_nanny.restart()



# -*- coding: utf-8 -*-

from .badelf import (
    PopulationAnalysis__VaspBader__BadelfMatproj,
    StaticEnergy__Vasp__PrebadelfMatproj,
)
from .bader import (
    PopulationAnalysis__VaspBader__BaderMatproj,
    StaticEnergy__Vasp__PrebaderMatproj,
)
from .elf import PopulationAnalysis__Vasp__ElfMatproj



# -*- coding: utf-8 -*-

import logging
from pathlib import Path

from simmate.toolkit import Composition
from simmate.toolkit.structure_prediction.evolution.database.variable_nsites_composition import (
    VariableNsitesCompositionSearch,
)
from simmate.toolkit.structure_prediction.evolution.workflows.fixed_composition import (
    StructurePrediction__Toolkit__FixedComposition,
)
from simmate.workflow_engine import Workflow


class StructurePrediction__Toolkit__VariableNsitesComposition(Workflow):
    """
    Runs an evolutionary search algorithm to predict the most stable structure
    of a specific composition but with variable number of sites in the unitcell.

    For example, this would be Ca2N and up to 12 atoms (Ca8N4).
    """

    database_table = VariableNsitesCompositionSearch

    fixed_comp_workflow = StructurePrediction__Toolkit__FixedComposition

    @classmethod
    def run_config(
        cls,
        composition: str | Composition,
        directory: Path = None,
        singleshot_sources: list[str] = [],
        **kwargs,  # passed to fixed_comp_workflow
    ):

        # Start by submitting the singleshot sources for each factor size.
        logging.warning("Single-shot sources are not implemented yet")

        # Now we are going to cycle through factor sizes, starting at the lowest
        # and working our way up. We do this because:
        #   1. smaller unitcells run much faster (for DFT-calculation)
        #   2. smaller compositions converge much faster (for number of
        #       calcs required)
        #   3. results from smaller unitcells can help to speed up the larger
        #       compositional searches.

        # The search will involve N steps, where N is based on the reduced
        # composition. For example, Ca6N3 would have a reduced comp of Ca2N
        # and max factor of 3.
        composition_reduced = composition.reduced_composition
        max_factor = int(composition.num_atoms / composition_reduced.num_atoms)

        # Starting at 1, go through each composition step and run a individual
        # fixed_comp_workflow for that composition.
        for factor in range(1, max_factor + 1):

            composition_current = composition_reduced * factor

            # logging.info(f"Beginning composition {composition_current}")
            cls.fixed_comp_workflow.run(
                composition=composition_current,
                # we keep the same base directory as we don't mind files overwriting
                # eachother as the search progresses through each factor
                directory=directory,
                # Because we submitted all steady states above, we don't
                # need the other workflows to do these anymore.
                singleshot_sources=[],
                **kwargs,
            )

            # logging.info(f"Completed composition {composition_current}")
